"""This module provides a unified API that allows other library modules to save acquired camera frames via the FFMPEG
library. Primarily, this module abstracts the configuration and flow control steps typically involved in saving acquired
camera frames as video files in real time.
"""

from enum import IntEnum, StrEnum
from typing import Any, ClassVar
from pathlib import Path
import subprocess
from subprocess import Popen, TimeoutExpired

import numpy as np
from numpy.typing import NDArray
from ataraxis_base_utilities import console, ensure_directory_exists


def check_gpu_availability() -> bool:
    """Checks whether the host system has an Nvidia GPU.

    The presence of a GPU is determined by calling the 'nvidia-smi' command. If the command runs successfully, it
    indicates the host-system, has an Nvidia GPU.

    Returns:
        True if the host system has an Nvidia GPU, False otherwise.
    """
    try:
        # Runs nvidia-smi command, uses check to trigger CalledProcessError exception if the runtime fails
        subprocess.run(
            args=["nvidia-smi", "--query-gpu=name", "--format=csv,noheader"],
            capture_output=True,
            text=True,
            check=True,
        )
    except Exception:
        return False
    else:
        return True


def check_ffmpeg_availability() -> bool:
    """Checks whether the host system has the FFMPEG library installed and available on PATH.

    The presence of the FFMPEG library is determined by calling the 'ffmpeg -version' command.

    Returns:
        True if the host system has the FFMPEG library installed and available on PATH, False otherwise.
    """
    try:
        # Runs ffmpeg version command, uses check to trigger CalledProcessError exception if runtime fails
        subprocess.run(args=["ffmpeg", "-version"], capture_output=True, text=True, check=True)
    except Exception:
        return False
    else:
        return True


class VideoEncoders(StrEnum):
    """Stores the supported video encoders used when saving camera frames as videos via VideoSaver instances."""

    H264 = "H264"
    """
    For CPU savers this is the libx264 encoder and for GPU savers this is the h264_nvenc encoder.
    """
    H265 = "H265"
    """
    For CPU savers this is the libx265 encoder and for GPU savers this is the hevc_nvenc encoder.
    """


class EncoderSpeedPresets(IntEnum):
    """Stores the supported video encoding speed presets used when saving camera frames as videos via VideoSaver
    instances.

    Generally, the faster the encoding speed, the lower is the resultant video quality.

    Notes:
        It is impossible to perfectly match the encoding presets for the CPU and GPU encoders. The scale defined
        in this enumeration represents the best-effort to align the preset scale for the two encoders.
    """

    FASTEST = 1
    """
    For CPU encoders, this matches the 'veryfast' level. For GPU encoders, this matches the 'p1' level.
    """
    FASTER = 2
    """
    For CPU encoders, this matches the 'faster' level. For GPU encoders, this matches the 'p2' level.
    """
    FAST = 3
    """
    For CPU encoders, this matches the 'fast' level. For GPU encoders, this matches the 'p3' level.
    """
    MEDIUM = 4
    """
    For CPU encoders, this matches the 'medium' level. For GPU encoders, this matches the 'p4' level.
    """
    SLOW = 5
    """
    For CPU encoders, this matches the 'slow' level. For GPU encoders, this matches the 'p5' level.
    """
    SLOWER = 6
    """
    For CPU encoders, this matches the 'slower' level. For GPU encoders, this matches the 'p6' level.
    """
    SLOWEST = 7
    """
    For CPU encoders, this matches the 'veryslow' level. For GPU encoders, this matches the 'p7' level.
    """


class InputPixelFormats(StrEnum):
    """Stores the supported camera frame data (color) formats used when saving camera frames as videos via VideoSaver
    instances.
    """

    MONOCHROME = "gray"
    """
    The preset for grayscale (monochrome) images.
    """
    BGR = "bgr24"
    """
    The preset for color images.
    """


class OutputPixelFormats(StrEnum):
    """Stores the supported video color formats used when saving camera frames as videos via VideoSaver instances."""

    YUV420 = "yuv420p"
    """
    The 'standard' video color space format that uses half-bandwidth chrominance (U/V) and full-bandwidth luminance (Y).
    Generally, the resultant reduction in chromatic precision is not apparent to the viewer.
    """
    YUV444 = "yuv444p"
    """
    While still minorly reducing the chromatic precision, this profile uses most of the chrominance channel-width. 
    This results in minimal chromatic data loss compared to the more common 'yuv420p' format, but increases the 
    encoding processing time.
    """


class VideoSaver:
    """Interfaces with an FFMPEG process to continuously save the input camera frames as an MP4 video file.

    This class uses the FFMPEG library and either Nvidia GPU or CPU to continuously encode and append the input stream
    of camera frames to an MP4 video file stored in non-volatile memory (on disk).

    Args:
        system_id: The unique identifier code of the VideoSystem instance that uses this saver interface.
        output_file: The path to the .mp4 video file to create at runtime.
        frame_width: The width of the video to be encoded, in pixels.
        frame_height: The height of the video to be encoded, in pixels.
        frame_rate: The frame rate of the video to be created.
        gpu: The index of the GPU to use for encoding. Setting this argument to a value of -1 (default) configures the
            instance to instead use the CPU for encoding. Valid GPU indices can be obtained from the 'nvidia-smi'
            terminal command.
        video_encoder: The encoder to use for generating the video file. Must be one of the valid VideoEncoders
            enumeration members.
        encoder_speed_preset: The encoding speed preset to use for generating the video file. Must be one of the valid
            EncoderSpeedPresets enumeration members.
        input_pixel_format: The pixel format used by the input frame data. This argument depends on the configuration of
            the camera used to acquire the frames. Must be one of the valid InputPixelFormats enumeration members.
        output_pixel_format: The pixel format to be used by the output video file. Must be one of the valid
            OutputPixelFormats enumeration members.
        quantization_parameter: The integer value to use for the 'quantization parameter' of the encoder. This
            determines how much information to discard from each encoded frame. Lower values produce better video
            quality at the expense of longer processing time and larger file size: 0 is best, 51 is worst. Note, the
            default value is calibrated for the H265 encoder and is likely too low for the H264 encoder.

    Attributes:
        _gpu_encoder_preset_map: Maps EncoderSpeedPresets enumeration member values to the appropriate GPU encoder
            speed preset values.
        _cpu_encoder_preset_map: Maps EncoderSpeedPresets enumeration member values to the appropriate CPU encoder
            speed preset values.
        _system_id: Stores the unique identifier code of the VideoSystem instance that uses this saver interface.
        _ffmpeg_command: Stores the main body of the FFMPEG command used to start the video encoding process.
        _repr_body: Stores the main body of the class representation string.
        _ffmpeg_process: Stores the Popen object that controls the FFMPEG's video encoding process. This is used during
            camera frame encoding to continuously feed the input camera frames to the encoding process.
    """

    _gpu_encoder_preset_map: ClassVar[dict[int, str]] = {
        1: "p1",
        2: "p2",
        3: "p3",
        4: "p4",
        5: "p5",
        6: "p6",
        7: "p7",
    }

    _cpu_encoder_preset_map: ClassVar[dict[int, str]] = {
        1: "veryfast",
        2: "faster",
        3: "fast",
        4: "medium",
        5: "slow",
        6: "slower",
        7: "veryslow",
    }

    def __init__(
        self,
        system_id: int,
        output_file: Path,
        frame_width: int,
        frame_height: int,
        frame_rate: float,
        gpu: int = -1,
        video_encoder: VideoEncoders | str = VideoEncoders.H265,
        encoder_speed_preset: EncoderSpeedPresets | int = EncoderSpeedPresets.SLOW,
        input_pixel_format: InputPixelFormats | str = InputPixelFormats.BGR,
        output_pixel_format: OutputPixelFormats | str = OutputPixelFormats.YUV420,
        quantization_parameter: int = 15,
    ) -> None:
        # Stores the caller VideoSystem ID to a class attribute.
        self._system_id: int = system_id

        # Ensures that all enumeration inputs are stored as enumerations:
        video_encoder = VideoEncoders(video_encoder)
        encoder_speed_preset = EncoderSpeedPresets(encoder_speed_preset)
        input_pixel_format = InputPixelFormats(input_pixel_format)
        output_pixel_format = OutputPixelFormats(output_pixel_format)

        # Ensures that the output file's directory exists.
        ensure_directory_exists(output_file)

        # Constructs the encoder-specific portion of the FFMPEG command based on whether GPU or CPU encoding is
        # requested. This portion contains the encoder parameters but lacks the input header and output path.
        encoder_command_portion: str

        # If a GPU index is provided, uses one of the hardware-encoding libraries.
        if gpu >= 0:
            # Selects the encoder library.
            video_codec = "h264_nvenc" if video_encoder == VideoEncoders.H264 else "hevc_nvenc"

            # Resolves the chromatic coding profile.
            if video_codec == "h264_nvenc":
                encoder_profile = "high444p" if output_pixel_format == OutputPixelFormats.YUV444 else "main"
            else:
                encoder_profile = "rext" if output_pixel_format == OutputPixelFormats.YUV444 else "main"

            # Resolves the GPU encoding speed preset.
            encoder_speed = self._gpu_encoder_preset_map[encoder_speed_preset.value]

            # Uses the resolved data to construct the GPU encoding command.
            encoder_command_portion = (
                f"-vcodec {video_codec} -qp {quantization_parameter} -preset {encoder_speed} "
                f"-profile:v {encoder_profile} -pixel_format {output_pixel_format.value} -gpu {gpu} -rc constqp"
            )

        # Otherwise, uses one of the software-encoding libraries.
        else:
            # Selects the encoder library.
            video_codec = "libx264" if video_encoder == VideoEncoders.H264 else "libx265"

            # Resolves the chromatic coding profile.
            if video_codec == "libx265":
                encoder_profile = "main444-8" if output_pixel_format == OutputPixelFormats.YUV444 else "main"
            else:
                encoder_profile = "high444" if output_pixel_format == OutputPixelFormats.YUV444 else "high422"

            # Resolves the CPU encoding speed preset.
            encoder_speed = self._cpu_encoder_preset_map[encoder_speed_preset.value]

            # This is unique to CPU encoders. Resolves the 'parameter' specifier based on the encoder name. This is
            # used to force CPU encoders to use the QP control mode.
            parameter_specifier = "-x264-params" if video_codec == "libx264" else "-x265-params"

            # Note, the qp has to be preceded by the '-parameter' specifier for the desired h265 / h265 encoder
            encoder_command_portion = (
                f"-vcodec {video_codec} {parameter_specifier} qp={quantization_parameter} "
                f"-preset {encoder_speed} -profile {encoder_profile} -pix_fmt "
                f"{output_pixel_format.value}"
            )

        # Constructs the complete FFMPEG command used by the start() method to initialize the encoder process. This
        # includes the input specifications, encoder parameters, and output path.
        self._ffmpeg_command: str = (
            f"ffmpeg -y -f rawvideo -pix_fmt {input_pixel_format.value} -s {frame_width}x{frame_height} "
            f"-r {frame_rate} -i pipe: {encoder_command_portion} {output_file}"
        )

        # Also generates the body for the representation string used by the __repr__() method. This is done here to
        # reduce the number of class attributes.
        self._repr_body: str = (
            f"output_file={output_file}, hardware_encoding={gpu >= 0}, "
            f"input_pixel_format={input_pixel_format.value}, video_encoder={video_encoder}, "
            f"encoding_speed_preset={encoder_speed_preset.value}, quantization_parameter={quantization_parameter}, "
            f"gpu_index={gpu}"
        )

        # Initializes the attribute to store the live FFMPEG encoder process, once it is started.
        self._ffmpeg_process: Popen[bytes] | None = None

    def __repr__(self) -> str:
        """Returns the string representation of the VideoEncoder instance."""
        return f"VideoSaver({self._repr_body}, started={self._ffmpeg_process is None})"

    def __del__(self) -> None:
        """Ensures that the video encoder is stopped before the instance is garbage-collected."""
        self.stop()

    @property
    def is_active(self) -> bool:
        """Returns True if the instance's encoder process is active (running)."""
        return self._ffmpeg_process is not None

    def start(self) -> None:
        """Creates the FFMPEG encoder process and sets up the data stream to pipe incoming camera frames to the
        process.
        """
        # Prevents recreating an already existing process
        if self._ffmpeg_process is not None:
            return

        # Starts the FFMPEG process using the command constructed during initialization and saves it to class attribute.
        self._ffmpeg_process = subprocess.Popen(
            self._ffmpeg_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        )

    def stop(self) -> None:
        """Stops the FFMPEG encoder process."""
        # Prevents stopping an already stopped process.
        if self._ffmpeg_process is None:
            return

        # If the process does not terminate 'gracefully,' it is terminated forcefully to prevent deadlocks.
        try:
            _ = self._ffmpeg_process.communicate(timeout=600)
        except TimeoutExpired:  # pragma: no cover
            self._ffmpeg_process.kill()

        # Sets the process variable to None placeholder. This causes the underlying Popen object to be garbage
        # collected.
        self._ffmpeg_process = None

    def save_frame(self, frame: NDArray[np.integer[Any]]) -> None:
        """Sends the input frame to be added to the video file managed by the instance's FFMPEG encoder process.

        Notes:
            This method expects that the input frame data matches the video dimensions and input pixel format used
            during VideoSaver initialization.

        Args:
            frame: The frame's data to be encoded into the video.

        Raises:
            ConnectionError: If the method is called before starting the encoder process via the start() method.
            BrokenPipeError: If the method encounters an error when submitting the frame's data to the FFMPEG process.
        """
        # Raises an error if the encoder process does not exist
        if self._ffmpeg_process is None:
            message = (
                f"Unable to submit the frame's data to the FFMPEG encoder process of the VideoSaver instance for the "
                f"VideoSystem with id {self._system_id} as the process has not been started. Call the start() method "
                f"to start the encoder process before calling the save_frame() method."
            )
            console.error(message=message, error=ConnectionError)

        # Writes the input frame to the encoder's standard input pipe.
        try:
            self._ffmpeg_process.stdin.write(frame.tobytes())  # type: ignore[union-attr]
        except Exception as e:  # pragma: no cover
            message = (
                f"The FFMPEG process of the VideoSaver instance for the VideoSystem with id {self._system_id} "
                f"has failed to process the input frame's data with error: {e}"
            )
            console.error(message=message, error=BrokenPipeError)



"""This example script demonstrates how to use the library to record and display the frames acquired by a personal
webcam.
See API documentation at https://ataraxis-video-system-api-docs.netlify.app/ for additional configuration options
exposed by VideoSystem interface.
See https://github.com/Sun-Lab-NBB/ataraxis-video-system?tab=readme-ov-file#usage for more details on how to use this
library.
Authors: Ivan Kondratyev (Inkaros), Jacob Groner, Natalie Yeung
"""

from pathlib import Path

import numpy as np
import tempfile
from ataraxis_time import PrecisionTimer
from ataraxis_data_structures import DataLogger, assemble_log_archives
from ataraxis_base_utilities import console, LogLevel

from ataraxis_video_system import VideoSystem, VideoEncoders, CameraInterfaces, extract_logged_camera_timestamps

# Since the VideoSystem and DataLogger classes use multiprocessing under-the-hood, the runtime must be protected by the
# __main__ guard.
if __name__ == "__main__":
    # Enables the console module to communicate the example's runtime progress via the terminal.
    console.enable()

    # Specifies the directory where to save the acquired video frames and timestamps.
    tempdir = tempfile.TemporaryDirectory()  # Creates a temporary directory for illustration purposes
    output_directory = Path(tempdir.name)

    # The DataLogger is used to save frame acquisition timestamps to disk as uncompressed .npy files.
    logger = DataLogger(output_directory=output_directory, instance_name="webcam")

    # The DataLogger has to be started before it can save any log entries.
    logger.start()

    # The VideoSystem minimally requires an ID and a DataLogger instance. The ID is critical, as it is used to identify
    # the log entries generated by the VideoSystem. For VideoSystems that will be saving frames, output_directory is
    # also required
    vs = VideoSystem(
        system_id=np.uint8(101),
        data_logger=logger,
        output_directory=output_directory,
        camera_interface=CameraInterfaces.OPENCV,  # OpenCV interface for webcameras
        display_frame_rate=15,  # Displays the acquired data at a rete of 15 frames per second
        color=False,  # Acquires images in MONOCHROME mode
        video_encoder=VideoEncoders.H264,  # Uses H264 CPU video encoder.
        quantization_parameter=25,  # Increments the default qp parameter to reflect using the H264 encoder.
    )

    # Calling this method arms the video system and starts frame acquisition. However, the frames are not initially
    # saved to disk.
    vs.start()
    console.echo(f"VideoSystem: Started", level=LogLevel.SUCCESS)

    console.echo(f"Acquiring frames without saving...")
    timer = PrecisionTimer("s")
    timer.delay(delay=5, block=False)  # During this delay, camera frames are displayed to the user but are not saved

    # Begins saving frames to disk as an MP4 video file
    console.echo(f"Saving the acquired frames to disk...")
    vs.start_frame_saving()
    timer.delay(delay=5, block=False)  # Records frames for 5 seconds, generating ~150 frames
    vs.stop_frame_saving()

    # Frame acquisition can be started and stopped as needed, although all frames are written to the same output
    # video file.

    # Stops the VideoSystem runtime and releases all resources
    vs.stop()
    console.echo(f"VideoSystem: Stopped", level=LogLevel.SUCCESS)

    # Stops the DataLogger and assembles all logged data into a single .npz archive file. This step is required to be
    # able to extract the timestamps for further analysis.
    logger.stop()
    console.echo(f"Assembling the frame timestamp log archive...")
    assemble_log_archives(remove_sources=True, log_directory=logger.output_directory, verbose=True)

    # Extracts the list of frame timestamps from the assembled log archive generated above. This returns a list of
    # timestamps. Each is given in microseconds elapsed since the UTC epoch onset.
    console.echo(f"Extracting frame acquisition timestamps from the assembled log archive...")
    timestamps = extract_logged_camera_timestamps(log_path=logger.output_directory.joinpath(f"101_log.npz"))

    # Computes and prints the frame rate of the camera based on the extracted frame timestamp data.
    timestamp_array = np.array(timestamps, dtype=np.uint64)
    time_diffs = np.diff(timestamp_array)
    fps = 1 / (np.mean(time_diffs) / 1e6)
    console.echo(
        message=(
            f"According to the extracted timestamps, the interfaced camera had an acquisition frame rate of "
            f"approximately {fps:.2f} frames / second."
        ),
        level=LogLevel.SUCCESS,
    )

    # Cleans up the temporary directory before shutting the runtime down.
    tempdir.cleanup()



"""Contains tests for classes and methods provided by the saver.py module."""

import subprocess

import numpy as np
import pytest
from ataraxis_base_utilities import error_format

from ataraxis_video_system.saver import (
    VideoSaver,
    VideoEncoders,
    EncoderSpeedPresets,
    InputPixelFormats,
    OutputPixelFormats,
    check_gpu_availability,
    check_ffmpeg_availability,
)
from ataraxis_video_system.camera import MockCamera


@pytest.fixture(scope="session")
def has_nvidia():
    """Checks for NVIDIA GPU availability in the test environment."""
    return check_gpu_availability()


@pytest.fixture(scope="session")
def has_ffmpeg():
    """Checks for FFMPEG availability in the test environment."""
    return check_ffmpeg_availability()


def test_check_gpu_availability():
    """Verifies the functioning of the check_gpu_availability() function."""
    # Tests that the function returns a boolean
    result = check_gpu_availability()
    assert isinstance(result, bool)

    # If nvidia-smi is available, verifies it returns True
    try:
        subprocess.run(
            ["nvidia-smi", "--query-gpu=name", "--format=csv,noheader"],
            capture_output=True,
            text=True,
            check=True,
            timeout=5,
        )
        assert result is True
    except:
        assert result is False


def test_check_ffmpeg_availability():
    """Verifies the functioning of the check_ffmpeg_availability() function."""
    # Tests that the function returns a boolean
    result = check_ffmpeg_availability()
    assert isinstance(result, bool)

    # If ffmpeg is available, verifies it returns True
    try:
        subprocess.run(
            ["ffmpeg", "-version"],
            capture_output=True,
            text=True,
            check=True,
            timeout=5,
        )
        assert result is True
    except:
        assert result is False


def test_video_saver_init_repr(tmp_path, has_ffmpeg):
    """Verifies the functioning of the VideoSaver __init__() and __repr__() methods."""
    if not has_ffmpeg:
        pytest.skip("Skipping this test as it requires FFMPEG.")

    # Tests CPU encoder initialization
    output_file = tmp_path / "test_video.mp4"
    saver = VideoSaver(
        system_id=1,
        output_file=output_file,
        frame_width=640,
        frame_height=480,
        frame_rate=30.0,
        gpu=-1,
        video_encoder=VideoEncoders.H265,
        encoder_speed_preset=EncoderSpeedPresets.MEDIUM,
        input_pixel_format=InputPixelFormats.BGR,
        output_pixel_format=OutputPixelFormats.YUV420,
        quantization_parameter=20,
    )

    # Verifies that the saver was initialized properly
    assert saver._system_id == 1
    assert saver._ffmpeg_process is None
    assert not saver.is_active  # Note: is_active returns True when the process is None

    # Verifies the __repr__() method
    assert "VideoSaver(" in repr(saver)
    assert "output_file=" in repr(saver)
    assert "hardware_encoding=False" in repr(saver)


@pytest.mark.parametrize(
    "video_encoder, gpu_index, output_pixel_format",
    [
        (VideoEncoders.H265, -1, OutputPixelFormats.YUV420),
        (VideoEncoders.H264, -1, OutputPixelFormats.YUV420),
        (VideoEncoders.H265, -1, OutputPixelFormats.YUV444),
        (VideoEncoders.H264, -1, OutputPixelFormats.YUV444),
    ],
)
def test_video_saver_cpu_configurations(tmp_path, video_encoder, gpu_index, output_pixel_format, has_ffmpeg):
    """Verifies different CPU encoder configurations for the VideoSaver class."""
    if not has_ffmpeg:
        pytest.skip("Skipping this test as it requires FFMPEG.")

    output_file = tmp_path / "test_video.mp4"
    saver = VideoSaver(
        system_id=1,
        output_file=output_file,
        frame_width=320,
        frame_height=240,
        frame_rate=15.0,
        gpu=gpu_index,
        video_encoder=video_encoder,
        encoder_speed_preset=EncoderSpeedPresets.FASTEST,
        input_pixel_format=InputPixelFormats.BGR,
        output_pixel_format=output_pixel_format,
        quantization_parameter=25,
    )

    # Verifies the FFMPEG command was constructed properly
    assert "libx264" in saver._ffmpeg_command or "libx265" in saver._ffmpeg_command
    # noinspection PyTypeChecker
    assert output_pixel_format.value in saver._ffmpeg_command
    assert "veryfast" in saver._ffmpeg_command  # FASTEST maps to veryfast for CPU


@pytest.mark.parametrize(
    "video_encoder, output_pixel_format",
    [
        (VideoEncoders.H265, OutputPixelFormats.YUV420),
        (VideoEncoders.H264, OutputPixelFormats.YUV420),
        (VideoEncoders.H265, OutputPixelFormats.YUV444),
        (VideoEncoders.H264, OutputPixelFormats.YUV444),
    ],
)
def test_video_saver_gpu_configurations(tmp_path, video_encoder, output_pixel_format, has_nvidia, has_ffmpeg):
    """Verifies different GPU encoder configurations for the VideoSaver class."""
    if not has_nvidia:
        pytest.skip("Skipping this test as it requires an NVIDIA GPU.")
    if not has_ffmpeg:
        pytest.skip("Skipping this test as it requires FFMPEG.")

    output_file = tmp_path / "test_video.mp4"
    saver = VideoSaver(
        system_id=1,
        output_file=output_file,
        frame_width=320,
        frame_height=240,
        frame_rate=15.0,
        gpu=0,
        video_encoder=video_encoder,
        encoder_speed_preset=EncoderSpeedPresets.FASTEST,
        input_pixel_format=InputPixelFormats.BGR,
        output_pixel_format=output_pixel_format,
        quantization_parameter=25,
    )

    # Verifies the FFMPEG command was constructed properly for GPU encoding
    assert "h264_nvenc" in saver._ffmpeg_command or "hevc_nvenc" in saver._ffmpeg_command
    # noinspection PyTypeChecker
    assert output_pixel_format.value in saver._ffmpeg_command
    assert "p1" in saver._ffmpeg_command  # FASTEST maps to p1 for GPU
    assert "-gpu 0" in saver._ffmpeg_command


def test_video_saver_start_stop(tmp_path, has_ffmpeg):
    """Verifies the functioning of the VideoSaver start() and stop() methods."""
    if not has_ffmpeg:
        pytest.skip("Skipping this test as it requires FFMPEG.")

    output_file = tmp_path / "test_video.mp4"
    saver = VideoSaver(
        system_id=1,
        output_file=output_file,
        frame_width=100,
        frame_height=100,
        frame_rate=10.0,
        gpu=-1,
        video_encoder=VideoEncoders.H265,
        encoder_speed_preset=EncoderSpeedPresets.FASTEST,
        input_pixel_format=InputPixelFormats.BGR,
        output_pixel_format=OutputPixelFormats.YUV420,
        quantization_parameter=30,
    )

    # Verifies that the process is not running initially
    assert saver._ffmpeg_process is None

    # Starts the encoder process
    saver.start()
    assert saver._ffmpeg_process is not None

    # Verifies that calling start() again does nothing
    process = saver._ffmpeg_process
    saver.start()
    assert saver._ffmpeg_process is process  # Same process object

    # Stops the encoder process
    saver.stop()
    assert saver._ffmpeg_process is None

    # Verifies that calling stop() again does nothing
    saver.stop()
    assert saver._ffmpeg_process is None


def test_video_saver_save_frame(tmp_path, has_ffmpeg):
    """Verifies the functioning of the VideoSaver save_frame() method."""
    if not has_ffmpeg:
        pytest.skip("Skipping this test as it requires FFMPEG.")

    # Setup
    output_file = tmp_path / "test_video.mp4"
    frame_width = 100
    frame_height = 100

    # Creates a mock camera to generate test frames
    camera = MockCamera(system_id=1, color=True, frame_rate=10, frame_width=frame_width, frame_height=frame_height)
    camera.connect()

    # Creates the video saver
    saver = VideoSaver(
        system_id=1,
        output_file=output_file,
        frame_width=frame_width,
        frame_height=frame_height,
        frame_rate=10.0,
        gpu=-1,
        video_encoder=VideoEncoders.H264,
        encoder_speed_preset=EncoderSpeedPresets.FASTEST,
        input_pixel_format=InputPixelFormats.BGR,
        output_pixel_format=OutputPixelFormats.YUV420,
        quantization_parameter=35,
    )

    # Starts the encoder
    saver.start()

    # Generates and saves test frames
    for _ in range(20):
        frame = camera.grab_frame()
        saver.save_frame(frame)

    # Stops the encoder to finalize the video
    saver.stop()

    # Verifies that the video file was created
    assert output_file.exists()
    assert output_file.stat().st_size > 0  # File is not empty


def test_video_saver_save_frame_errors(tmp_path, has_ffmpeg):
    """Verifies the error handling of the VideoSaver save_frame() method."""
    if not has_ffmpeg:
        pytest.skip("Skipping this test as it requires FFMPEG.")

    output_file = tmp_path / "test_video.mp4"
    saver = VideoSaver(
        system_id=1,
        output_file=output_file,
        frame_width=100,
        frame_height=100,
        frame_rate=10.0,
        gpu=-1,
    )

    # Creates a test frame
    frame = np.zeros((100, 100, 3), dtype=np.uint8)

    # Verifies that saving a frame without starting the encoder raises an error
    message = (
        f"Unable to submit the frame's data to the FFMPEG encoder process of the VideoSaver instance for the "
        f"VideoSystem with id 1 as the process has not been started. Call the start() method "
        f"to start the encoder process before calling the save_frame() method."
    )
    with pytest.raises(ConnectionError, match=error_format(message)):
        saver.save_frame(frame)


def test_video_saver_del(tmp_path, has_ffmpeg):
    """Verifies that the VideoSaver __del__() method properly cleans up resources."""
    if not has_ffmpeg:
        pytest.skip("Skipping this test as it requires FFMPEG.")

    output_file = tmp_path / "test_video.mp4"
    saver = VideoSaver(
        system_id=1,
        output_file=output_file,
        frame_width=100,
        frame_height=100,
        frame_rate=10.0,
        gpu=-1,
    )

    # Starts the encoder
    saver.start()
    assert saver._ffmpeg_process is not None

    # Deletes the saver (should call stop() internally)
    del saver

    # Creates a new saver to verify resources were released
    saver2 = VideoSaver(
        system_id=1,
        output_file=output_file,
        frame_width=100,
        frame_height=100,
        frame_rate=10.0,
        gpu=-1,
    )
    # Should be able to start without conflicts
    saver2.start()
    saver2.stop()


def test_encoder_speed_preset_mappings():
    """Verifies that the encoder speed preset mappings are correctly defined."""
    # Verifies all EncoderSpeedPresets values have corresponding mappings
    for preset in EncoderSpeedPresets:
        assert preset.value in VideoSaver._gpu_encoder_preset_map
        assert preset.value in VideoSaver._cpu_encoder_preset_map

    # Verifies the specific mappings
    assert VideoSaver._gpu_encoder_preset_map[EncoderSpeedPresets.FASTEST] == "p1"
    assert VideoSaver._gpu_encoder_preset_map[EncoderSpeedPresets.SLOWEST] == "p7"
    assert VideoSaver._cpu_encoder_preset_map[EncoderSpeedPresets.FASTEST] == "veryfast"
    assert VideoSaver._cpu_encoder_preset_map[EncoderSpeedPresets.SLOWEST] == "veryslow"



"""This module provides a unified API that allows other library modules to interface with any supported camera hardware.
Primarily, these interfaces abstract the necessary procedures to connect to the camera and continuously grab the
acquired frames.
"""

from enum import StrEnum
from typing import Any
from pathlib import Path
from dataclasses import dataclass

import cv2
import numpy as np
import appdirs
from numpy.typing import NDArray
from ataraxis_time import PrecisionTimer
from harvesters.core import Harvester, ImageAcquirer  # type: ignore[import-untyped]
from harvesters.util.pfnc import (  # type: ignore[import-untyped]
    bgr_formats,
    rgb_formats,
    mono_location_formats,
)
from ataraxis_base_utilities import console, ensure_directory_exists

from .saver import InputPixelFormats

# Repackages Harvesters color formats into sets to optimize the efficiency of the HarvestersCamera grab_frames() method:
_mono_formats = set(mono_location_formats)
_color_formats = set(bgr_formats) | set(rgb_formats)
_all_rgb_formats = set(rgb_formats)

# Determines the size of the frame pool used by the MockCamera instances.
_FRAME_POOL_SIZE = 10

# Determines the maximum number of failed test attempts allowed when running the get_opencv_ids() function before the
# runtime is terminated. This is used to interrupt the function early when it runs out of valid OpenCV-compatible camera
# objects to test.
_MAXIMUM_NON_WORKING_IDS = 5


class CameraInterfaces(StrEnum):
    """Specifies the supported camera interface backends compatible with the VideoSystem class."""

    HARVESTERS = "harvesters"
    """
    This is the preferred backend for all cameras that support the GeniCam standard, which includes most scientific and
    industrial machine-vision cameras. This backend is based on the 'Harvesters' library and works with all 
    GeniCam-compatible interfaces (USB, Ethernet, PCIE).
    """
    OPENCV = "opencv"
    """
    This is the backend used for all cameras that do not support the GeniCam standard. This backend is based on the 
    'OpenCV' library and primarily works for consumer-grade cameras that use the USB interface.
    """
    MOCK = "mock"
    """
    This backend is used exclusively for internal library testing and should not be used in production projects.
    """


@dataclass()
class CameraInformation:
    """Stores descriptive information about a camera discoverable through OpenCV or Harvesters libraries."""

    camera_index: int
    """The index of the camera in the list of all cameras discoverable through the evaluated interface 
    (OpenCV or Harvesters)."""
    interface: CameraInterfaces | str
    """The interface that discovered the camera."""
    frame_width: int
    """The width of the frames acquired by the camera, in pixels."""
    frame_height: int
    """The height of the frames acquired by the camera, in pixels."""
    acquisition_frame_rate: int
    """The frame rate at which the camera acquires frames, in frames per second."""
    serial_number: str | None = None
    """Only for Harvesters-discoverable cameras. Contains the camera's serial number."""
    model: str | None = None
    """Only for Harvesters-discoverable cameras. Contains the camera's model name."""


def get_opencv_ids() -> tuple[CameraInformation, ...]:
    """Discovers and reports the identifier (indices) and descriptive information about the cameras accessible through
    the OpenCV library.

    Notes:
        Currently, it is impossible to retrieve serial numbers or camera models from OpenCV. Therefore, while this
        method tries to provide some ID information, it is typically insufficient to identify specific cameras. It is
        advised to test each discovered camera with the 'axvs run' CLI command to identify the mapping between the
        discovered indices (IDs) and physical cameras.

    Returns:
         A tuple of CameraData instances, one for each discovered OpenCV-compatible camera.
    """
    # Disables OpenCV error logging to avoid flushing the terminal with failed connection attempts.
    prev_log_level = cv2.getLogLevel()
    cv2.setLogLevel(0)

    try:
        non_working_count = 0
        working_ids: list[CameraInformation] = []

        # This loop iterates over IDs until it discovers 5 non-working IDs. The loop is designed to evaluate 100 IDs at
        # maximum to prevent infinite execution.
        for evaluated_id in range(100):
            # Evaluates each ID (index) by instantiating a video-capture object and reading one image and dimension data
            # from the connected camera (if any was connected).
            camera = cv2.VideoCapture(evaluated_id)

            # If the evaluated camera can be connected and returns images, it's index is appended to the ID list
            if camera.isOpened() and camera.read()[0]:
                frame_width = int(camera.get(cv2.CAP_PROP_FRAME_WIDTH))
                frame_height = int(camera.get(cv2.CAP_PROP_FRAME_HEIGHT))
                acquisition_rate = int(camera.get(cv2.CAP_PROP_FPS))
                camera_data = CameraInformation(
                    camera_index=evaluated_id,
                    interface=CameraInterfaces.OPENCV,
                    frame_width=frame_width,
                    frame_height=frame_height,
                    acquisition_frame_rate=acquisition_rate,
                )
                working_ids.append(camera_data)
                non_working_count = 0  # Resets non-working count whenever a working camera is found.
            else:
                non_working_count += 1

            camera.release()  # Releases the camera object to recreate it above for the next cycle

            # Breaks the loop early if more than 5 non-working IDs are found consecutively
            if non_working_count >= _MAXIMUM_NON_WORKING_IDS:
                break

        return tuple(working_ids)  # Converts to tuple before returning to caller.

    finally:
        # Restores previous log level
        cv2.setLogLevel(prev_log_level)


def get_harvesters_ids() -> tuple[CameraInformation, ...]:
    """Discovers and reports the identifier (indices) and descriptive information about the cameras accessible
    through the Harvesters library.

    Notes:
        This method bundles the discovered ID (index) information with the serial number and the model for each camera
        to support identifying the cameras.

    Returns:
        A tuple of CameraInformation instances, one for each discovered Harvesters-compatible camera.
    """
    # Instantiates the class and adds the input .cti file.
    harvester = Harvester()
    harvester.add_file(file_path=str(_get_cti_path()))

    # Gets the list of accessible cameras
    harvester.update()

    # Loops over all discovered cameras and retrieves detailed information from each camera
    working_ids: list[CameraInformation] = []
    for index, camera_info in enumerate(harvester.device_info_list):
        # Accesses the remote device node map to get camera properties
        camera = harvester.create(search_key=index)
        node_map = camera.remote_device.node_map

        # Retrieves frame dimensions and acquisition rate from the camera's node map.
        frame_width = int(node_map.Width.value)
        frame_height = int(node_map.Height.value)
        acquisition_rate = int(round(number=node_map.AcquisitionFrameRate.value, ndigits=0))

        # Creates CameraInformation instance with all retrieved data
        camera_data = CameraInformation(
            camera_index=index,  # Uses the enumerated index as the camera index
            interface=CameraInterfaces.HARVESTERS,
            frame_width=frame_width,
            frame_height=frame_height,
            acquisition_frame_rate=acquisition_rate,
            serial_number=camera_info.serial_number,
            model=camera_info.model,
        )
        working_ids.append(camera_data)

    # Resets the harvester instance after discovering the camera IDs.
    harvester.remove_file(file_path=str(_get_cti_path()))
    harvester.reset()

    return tuple(working_ids)  # Converts to tuple before returning to caller.


def add_cti_file(cti_path: Path) -> None:
    """Configures the 'harvesters' camera interface to use the provided .cti file during all future runtimes.

    The 'harvesters' camera interface requires the GenTL Producer interface (.cti) file to discover and interface with
    compatible GenTL devices (cameras). This function configures the local machine to use the specified .cti file
    for all future runtimes that use the 'harvesters' camera interface.

    Notes:
        The path to the .cti file is stored inside the user's data directory, so that it can be reused between library
        calls.

    Args:
        cti_path: The path to the CTI file that provides the GenTL Producer interface. It is recommended to use the
            file supplied by the camera vendor, but a general Producer, such as mvImpactAcquire, is also acceptable.
            See https://github.com/genicam/harvesters/blob/master/docs/INSTALL.rst for more details.
    """
    # Verifies the input CTI file.
    harvester = Harvester()
    harvester.add_file(file_path=str(cti_path), check_existence=True, check_validity=True)

    # Resolves the path to the library-specific .txt file used to store the path to the currently used .cti file.
    app_dir = Path(appdirs.user_data_dir(appname="ataraxis_video_system", appauthor="sun_lab"))
    cti_path_file = app_dir.joinpath("cti_path.txt")

    # In case this function is called before the app directory is created, ensures the app directory exists
    ensure_directory_exists(cti_path_file)

    # Overwrites the contents of the path file with the verified .cti file path.
    with cti_path_file.open("w") as f:
        f.write(str(cti_path))


def _get_cti_path() -> Path:
    """Resolves and returns the path to the CTI file that provides the GenTL Producer interface.

    This service function is used when initializing HarvestersCamera instances to resolve the GenTL Producer interface.

    Returns:
        The path to the GenTL Producer interface (.cti) file.

    Raises:
        FileNotFoundError: If the function is unable to resolve the path to the .cti file.
    """
    # Uses appdirs to locate the user's data directory and resolve the path to the .cti path file.
    app_dir = Path(appdirs.user_data_dir(appname="ataraxis_video_system", appauthor="sun_lab"))
    cti_path_file = app_dir.joinpath("cti_path.txt")

    # If the path file or the library data directory does not exist, aborts with an error.
    if not cti_path_file.exists():
        message = (
            "Unable to resolve the path to the GenTL Producer interface (.cti) file to use for the harvesters camera "
            "interface, as the .cti file has not been set. Set the .cti file path by calling the 'axvs cti' CLI "
            "command."
        )
        console.error(message=message, error=FileNotFoundError)

    # Once the location of the path storage file is resolved, reads the .cti file path.
    with cti_path_file.open() as f:
        cti_path = Path(f.read().strip())

    # Verifies the CTI file's validity before returning it to the caller.
    harvester = Harvester()
    harvester.add_file(file_path=str(cti_path), check_existence=True, check_validity=True)
    return cti_path


class OpenCVCamera:
    """Interfaces with the specified OpenCV-compatible camera hardware to acquire frame data.

    Notes:
        This class should not be initialized manually! Use the VideoSystem's add_camera() method to create all camera
        interface instances.

    Args:
        system_id: The unique identifier code of the VideoSystem instance that uses this camera interface.
        color: Specifies whether the camera acquires colored or monochrome images. This determines how to store the
            acquired frames. Colored frames are saved using the 'BGR' channel order, monochrome images are reduced to
            a single-channel format.
        camera_index: The index of the camera in the list of all cameras discoverable by OpenCV, e.g.: 0 for the first
            available camera, 1 for the second, etc. This specifies the camera hardware the instance should interface
            with at runtime.
        frame_rate: The desired rate, in frames per second, at which to capture the data. Note; whether the requested
            rate is attainable depends on the hardware capabilities of the camera and the communication interface. If
            this argument is not explicitly provided, the instance uses the default frame rate of the connected camera.
        frame_width: The desired width of the acquired frames, in pixels. Note; the requested width must be compatible
            with the range of frame dimensions supported by the camera hardware. If this argument is not explicitly
            provided, the instance uses the default frame width of the connected camera.
        frame_height: Same as 'frame_width', but specifies the desired height of the acquired frames, in pixels. If this
            argument is not explicitly provided, the instance uses the default frame width of the connected camera.

    Attributes:
        _system_id: Stores the unique identifier code of the VideoSystem instance that uses this camera interface.
        _color: Specifies whether the camera acquires colored or monochrome images.
        _camera_index: Stores the index of the camera hardware in the list of all OpenCV-discoverable cameras connected
            to the host-machine.
        _frame_rate: Stores the camera's frame acquisition rate.
        _frame_width: Stores the width of the camera's frames.
        _frame_height: Stores the height of the camera's frames.
        _camera: Stores the OpenCV VideoCapture object that interfaces with the camera.
        _acquiring: Tracks whether the camera is currently acquiring frames.
    """

    def __init__(
        self,
        system_id: int,
        camera_index: int = 0,
        frame_rate: int | None = None,
        frame_width: int | None = None,
        frame_height: int | None = None,
        *,
        color: bool = True,
    ) -> None:
        # Saves class parameters to class attributes
        self._system_id: int = system_id
        self._color: bool = color
        self._camera_index: int = camera_index
        self._frame_rate: int = 0 if frame_rate is None else frame_rate
        self._frame_width: int = 0 if frame_width is None else frame_width
        self._frame_height: int = 0 if frame_height is None else frame_height
        self._camera: cv2.VideoCapture | None = None
        self._acquiring: bool = False

    def __del__(self) -> None:
        """Releases the underlying VideoCapture object when the instance is garbage-collected."""
        self.disconnect()

    def __repr__(self) -> str:
        """Returns the string representation of the OpenCVCamera instance."""
        return (
            f"OpenCVCamera(system_id={self._system_id}, camera_index={self._camera_index}, "
            f"frame_rate={self.frame_rate} frames / second, frame_width={self.frame_width} pixels, "
            f"frame_height={self.frame_height} pixels, connected={self._camera is not None}, "
            f"acquiring={self._acquiring})"
        )

    def connect(self) -> None:
        """Connects to the managed camera hardware.

        Raises:
            ValueError: If the instance is configured to override hardware-defined acquisition parameters and the
                camera rejects the user-defined frame height, width, or acquisition rate parameters.
        """
        # Prevents re-connecting to an already connected camera
        if self._camera is not None:
            return

        # Instantiates the OpenCV VideoCapture object to acquire images from the camera, using the specified camera ID
        # (index).
        self._camera = cv2.VideoCapture(index=self._camera_index, apiPreference=cv2.CAP_ANY)

        # If necessary, overrides the requested camera acquisition parameters. If the camera does not accept the
        # requested parameters, terminates with an error message. Otherwise, queries the acquisition parameters from
        # the connected camera.
        if self._frame_rate != 0:
            self._camera.set(propId=cv2.CAP_PROP_FPS, value=float(self._frame_rate))
            actual_frame_rate = int(self._camera.get(propId=cv2.CAP_PROP_FPS))
            if actual_frame_rate < self._frame_rate:
                message = (
                    f"Unable to configure the OpenCVCamera interface for the VideoSystem with id {self._system_id}. "
                    f"Attempted configuring the camera to acquire frames at the rate of {self._frame_rate} "
                    f"frames per second, but the camera automatically adjusted the acquisition rate to "
                    f"{actual_frame_rate}. This indicates that the camera does not support the requested frame "
                    f"acquisition rate."
                )
                console.error(error=ValueError, message=message)
        else:
            self._frame_rate = int(self._camera.get(propId=cv2.CAP_PROP_FPS))

        if self._frame_width != 0:
            self._camera.set(propId=cv2.CAP_PROP_FRAME_WIDTH, value=float(self._frame_width))
            actual_frame_width = int(self._camera.get(propId=cv2.CAP_PROP_FRAME_WIDTH))
            if actual_frame_width != self._frame_width:
                message = (
                    f"Unable to configure the OpenCVCamera interface for the VideoSystem with id {self._system_id}. "
                    f"Attempted configuring the camera to acquire frames with the width of {self._frame_width} pixels, "
                    f"but the camera automatically adjusted the frame width to {actual_frame_width}. This indicates "
                    f"that the camera does not support the requested frame height and width combination."
                )
                console.error(error=ValueError, message=message)
        else:
            self._frame_width = int(self._camera.get(propId=cv2.CAP_PROP_FRAME_WIDTH))

        if self._frame_height != 0:
            self._camera.set(propId=cv2.CAP_PROP_FRAME_HEIGHT, value=float(self._frame_height))
            actual_frame_height = int(self._camera.get(propId=cv2.CAP_PROP_FRAME_HEIGHT))
            if actual_frame_height != self._frame_height:
                message = (
                    f"Unable to configure the OpenCVCamera interface for the VideoSystem with id {self._system_id}. "
                    f"Attempted configuring the camera to acquire frames with the height of {self._frame_height} "
                    f"pixels, but the camera automatically adjusted the frame height to {actual_frame_height}. This "
                    f"indicates that the camera does not support the requested frame height and width combination."
                )
                console.error(error=ValueError, message=message)
        else:
            self._frame_height = int(self._camera.get(propId=cv2.CAP_PROP_FRAME_HEIGHT))

    def disconnect(self) -> None:
        """Disconnects from the managed camera hardware."""
        # Prevents disconnecting from an already disconnected camera
        if self._camera is None:
            return

        # Disconnects from the camera
        self._camera.release()
        self._acquiring = False
        self._camera = None

    @property
    def is_connected(self) -> bool:
        """Returns True if the instance is connected to the camera hardware."""
        return self._camera is not None

    @property
    def is_acquiring(self) -> bool:
        """Returns True if the camera is currently acquiring video frames."""
        return self._acquiring

    @property
    def frame_rate(self) -> int:
        """Returns the acquisition rate of the camera, in frames per second (fps)."""
        return self._frame_rate

    @property
    def frame_width(self) -> int:
        """Returns the width of the acquired frames, in pixels."""
        return self._frame_width

    @property
    def frame_height(self) -> int:
        """Returns the height of the acquired frames, in pixels."""
        return self._frame_height

    @property
    def pixel_color_format(self) -> InputPixelFormats:
        """Returns the pixel color format of the acquired frames."""
        if self._color:
            return InputPixelFormats.BGR
        return InputPixelFormats.MONOCHROME

    def grab_frame(self) -> NDArray[np.floating[Any] | np.integer[Any]]:
        """Grabs the first available frame from the managed camera's acquisition buffer.

        This method has to be called repeatedly (cyclically) to fetch the newly acquired frames from the camera.

        Notes:
            The first time this method is called, the camera initializes frame acquisition, which is carried out
            asynchronously. If the camera supports buffering, it continuously saves the frames into its circular buffer.
            If the camera does not support buffering, the frame data must be fetched before the camera acquires the next
            frame to prevent frame loss.

            Due to the initial setup of the buffering procedure, the first call to this method incurs a significant
            delay.

        Returns:
            A NumPy array that stores the frame data. Depending on whether the camera acquires colored or monochrome
            images, the returned arrays have the shape (height, width, channels) or (height, width). Color data uses
            the BGR channel order.

        Raises:
            ConnectionError: If the instance is not connected to the camera hardware.
            BrokenPipeError: If the instance fails to fetch a frame from the connected camera hardware.
        """
        # Prevents calling this method before connecting to the camera's hardware
        if self._camera is None:
            message = (
                f"The OpenCVCamera instance for the VideoSystem with id {self._system_id} is not connected to the "
                f"camera hardware, and cannot acquire images. Call the connect() method prior to calling the "
                f"grab_frame() method."
            )
            console.error(message=message, error=ConnectionError)
            # Fallback to appease mypy, should not be reachable
            raise ConnectionError(message)  # pragma: no cover

        # Flips the acquisition tracker to True the first time this method is called for a connected camera.
        if not self._acquiring:
            self._acquiring = True

        frame: NDArray[np.floating[Any] | np.integer[Any]]
        success, frame = self._camera.read()
        if not success:
            message = (
                f"The OpenCVCamera instance for the VideoSystem with id {self._system_id} has failed to grab a frame "
                f"image from the camera hardware, which is not expected. This indicates initialization or connectivity "
                f"issues."
            )
            console.error(message=message, error=BrokenPipeError)

        if not self._color:
            # Converts the frame data from using BGR color space (default for all frames) to Monochrome if needed
            frame = cv2.cvtColor(src=frame, code=cv2.COLOR_BGR2GRAY)

        return frame


class HarvestersCamera:
    """Interfaces with the specified GeniCam-compatible camera hardware to acquire frame data.

    Notes:
        This class should not be initialized manually! Use the VideoSystem's add_camera() method to create all camera
        interface instances.

    Args:
        system_id: The unique identifier code of the VideoSystem instance that uses this camera interface.
        camera_index: The index of the camera in the list of all cameras discoverable by Harvesters, e.g.: 0 for the
            first available camera, 1 for the second, etc. This specifies the camera hardware the instance should
            interface with at runtime.
        frame_rate: The desired rate, in frames per second, at which to capture the data. Note; whether the requested
            rate is attainable depends on the hardware capabilities of the camera and the communication interface. If
            this argument is not explicitly provided, the instance uses the default frame rate of the connected camera.
        frame_width: The desired width of the acquired frames, in pixels. Note; the requested width must be compatible
            with the range of frame dimensions supported by the camera hardware. If this argument is not explicitly
            provided, the instance uses the default frame width of the connected camera.
        frame_height: Same as 'frame_width', but specifies the desired height of the acquired frames, in pixels. If this
            argument is not explicitly provided, the instance uses the default frame width of the connected camera.

    Attributes:
        _system_id: Stores the unique identifier code of the VideoSystem instance that uses this camera interface.
        _camera_index: Stores the index of the camera hardware in the list of all OpenCV-discoverable cameras connected
            to the host-machine.
        _frame_rate: Stores the camera's frame acquisition rate.
        _frame_width: Stores the width of the camera's frames.
        _frame_height: Stores the height of the camera's frames.
        _harvester: Stores the Harvester interface object that discovers and manages the list of accessible GenTL
            cameras.
        _camera: Stores the Harvesters ImageAcquirer object that interfaces with the camera.
        _color: Tracks whether the frames are acquired using a monochrome or a colored data format.
    """

    def __init__(
        self,
        system_id: int,
        camera_index: int = 0,
        frame_rate: int | None = None,
        frame_width: int | None = None,
        frame_height: int | None = None,
    ) -> None:
        # No input checking here as it is assumed that the class is initialized via get_camera() function that performs
        # the necessary input filtering.

        # Saves class parameters to class attributes
        self._system_id: int = system_id
        self._camera_index: int = camera_index
        self._frame_rate: int = 0 if frame_rate is None else frame_rate
        self._frame_width: int = 0 if frame_width is None else frame_width
        self._frame_height: int = 0 if frame_height is None else frame_height

        # Initializes the Harvester class to discover the list of available cameras.
        self._harvester: Harvester = Harvester()
        # Adds the .cti file to the class. This also verifies the file's existence and validity.
        self._harvester.add_file(file_path=str(_get_cti_path()), check_existence=True, check_validity=True)
        self._harvester.update()  # Discovers compatible cameras using the GenTL interface specified by the CTI file.

        # Pre-creates the attribute to store the initialized ImageAcquirer object for the connected camera.
        self._camera: ImageAcquirer | None = None

        # Tracks whether the acquired frames use a monochrome or a colored data format.
        self._color: bool = False

    def __del__(self) -> None:
        """Releases the underlying ImageAcquirer object when the instance is garbage-collected."""
        self.disconnect()  # Releases the camera object
        self._harvester.reset()  # Releases the Harvester class resources

    def __repr__(self) -> str:
        """Returns the string representation of the HarvestersCamera instance."""
        return (
            f"HarvestersCamera(system_id={self._system_id}, camera_index={self._camera_index}, "
            f"frame_rate={self.frame_rate} frames / second, frame_width={self.frame_width} pixels, "
            f"frame_height={self.frame_height} pixels, connected={self._camera is not None}, "
            f"acquiring={self.is_acquiring})"
        )

    def connect(self) -> None:
        """Connects to the managed camera hardware."""
        # Prevents connecting to an already connected camera.
        if self._camera is not None:
            return

        # Initializes an ImageAcquirer camera interface object to interface with the camera's hardware.
        self._camera = self._harvester.create(search_key=self._camera_index)

        # If necessary, overrides the requested camera acquisition parameters. Note, there is no guarantee that the
        # camera accepts the requested parameters.
        if self._frame_width != 0:
            self._camera.remote_device.node_map.Width.value = self._frame_width
        if self._frame_height != 0:
            self._camera.remote_device.node_map.Height.value = self._frame_height
        # The frame rate has to be set last, as it is affected by frame width and height
        if self._frame_rate != 0:
            self._camera.remote_device.node_map.AcquisitionFrameRate.value = self._frame_rate

        # Queries the current camera acquisition parameters and stores them in class attributes.
        self._frame_rate = int(self._camera.remote_device.node_map.AcquisitionFrameRate.value)
        self._frame_width = int(self._camera.remote_device.node_map.Width.value)
        self._frame_height = int(self._camera.remote_device.node_map.Height.value)

    def disconnect(self) -> None:
        """Disconnects from the managed camera hardware."""
        # Precents disconnecting from an already disconnected camera.
        if self._camera is None:
            return

        self._camera.stop()  # Stops image acquisition

        # Discards any unconsumed buffers to ensure proper memory release
        while self._camera.num_holding_filled_buffers != 0:
            _ = self._camera.fetch()  # pragma: no cover

        self._camera.destroy()  # Releases the camera object
        self._camera = None  # Sets the camera object to None

    @property
    def is_connected(self) -> bool:
        """Returns True if the instance is connected to the camera hardware."""
        return self._camera is not None

    @property
    def is_acquiring(self) -> bool:
        """Returns True if the camera is currently acquiring video frames."""
        if self._camera is not None:
            return bool(self._camera.is_acquiring())
        return False  # If the camera is not connected, it cannot be acquiring images.

    @property
    def frame_rate(self) -> int:
        """Returns the acquisition rate of the camera, in frames per second (fps)."""
        return self._frame_rate

    @property
    def frame_width(self) -> int:
        """Returns the width of the acquired frames, in pixels."""
        return self._frame_width

    @property
    def frame_height(self) -> int:
        """Returns the height of the acquired frames, in pixels."""
        return self._frame_height

    @property
    def pixel_color_format(self) -> InputPixelFormats:
        """Returns the pixel color format of the acquired frames."""
        if self._color:
            return InputPixelFormats.BGR
        return InputPixelFormats.MONOCHROME

    def grab_frame(self) -> NDArray[np.integer[Any]]:
        """Grabs the first available frame from the managed camera's acquisition buffer.

        This method has to be called repeatedly (cyclically) to fetch the newly acquired frames from the camera.

        Notes:
            The first time this method is called, the camera initializes frame acquisition, which is carried out
            asynchronously. The acquired frames are temporarily stored in the camera's circular buffer until they are
            fetched by this method.

            Due to the initial setup of the buffering procedure, the first call to this method incurs a significant
            delay.

        Returns:
            A NumPy array that stores the frame data. Depending on whether the camera acquires colored or monochrome
            images, the returned arrays have the shape (height, width, channels) or (height, width). Color data uses
            the BGR channel order.

        Raises:
            ConnectionError: If the instance is not connected to the camera hardware.
            BrokenPipeError: If the instance fails to fetch a frame from the connected camera hardware.
            ValueError: If the acquired frame data uses an unsupported data (color) format.
        """
        if not self._camera:
            message = (
                f"The HarvestersCamera instance for the VideoSystem with id {self._system_id} is not connected to the "
                f"camera hardware and cannot acquire images. Call the connect() method prior to calling the "
                f"grab_frame() method."
            )
            console.error(message=message, error=ConnectionError)
            # Fallback to appease mypy, should not be reachable
            raise ConnectionError(message)  # pragma: no cover

        # Triggers camera frame acquisition the first time this method is called.
        if not self._camera.is_acquiring():
            self._camera.start()

        # Retrieves the next available image buffer from the camera. Uses the 'with' context to properly
        # re-queue the buffer to acquire further images.
        with self._camera.fetch() as buffer:
            if buffer is None:  # pragma: no cover
                message = (
                    f"The HarvestersCamera instance for the VideoSystem with id {self._system_id} has failed to grab "
                    f"a frame image from the camera hardware, which is not expected. This indicates initialization or "
                    f"connectivity issues."
                )
                console.error(message=message, error=BrokenPipeError)

            # Retrieves the contents (frame data) from the buffer
            content = buffer.payload.components[0]

            # Collects the information necessary to reshape the originally 1-dimensional frame array into the
            # 2-dimensional array using the correct number and order of color channels.
            width = content.width
            height = content.height
            data_format = content.data_format

            # For monochrome formats, reshapes the 1D array into a 2D array and returns it to caller.
            if data_format in mono_location_formats:
                # Uses copy, which is VERY important. Once the buffer is released, the original 'content' is lost,
                # so NumPy needs to copy the data instead of using the default referencing behavior.
                out_array: NDArray[np.integer[Any]] = content.data.reshape(height, width).copy()
                self._color = False  # Ensures that the color flag is set to False.
                return out_array

            # For color data, evaluates the input format and reshapes the data as necessary.
            if data_format in _color_formats:  # pragma: no cover
                # Reshapes the data into RGB + A format as the first processing step.
                content.data.reshape(
                    height,
                    width,
                    int(content.num_components_per_pixel),  # Sets of R, G, B, and Alpha
                )

                # Swaps every R and B value (RGB → BGR) ot produce BGR / BGRA images. This ensures consistency
                # with the OpenCVCamera API. Note, this is only done if the image data is in the RGB format.
                if data_format in _all_rgb_formats:
                    frame: NDArray[np.integer[Any]] = content[:, :, ::-1].copy()

                self._color = True  # Ensures that the color flag is set to True.

                # Returns the reshaped frame array to the caller
                return frame

            # If the image has an unsupported data format, raises an error
            message = (
                f"The HarvestersCamera instance for the VideoSystem with id {self._system_id} has acquired an image "
                f"with an unsupported data (color) format {data_format}. Currently, only the following unpacked "
                f"families of color formats are supported: Monochrome, RGB, RGBA, BGR, and BGRA."
            )  # pragma: no cover
            console.error(message=message, error=ValueError)  # pragma: no cover
            # This should never be reached, it is here to appease mypy
            raise RuntimeError(ValueError)  # pragma: no cover


class MockCamera:
    """Simulates (mocks) the behavior of the OpenCVCamera and HarvestersCamera classes without the need to interface
    with a physical camera.

    This class is primarily used to test the VideoSystem class functionality. The class fully mimics the behavior of
    other camera interface classes but does not establish a physical connection with any camera hardware.

    Notes:
        This class should not be initialized manually! Use the VideoSystem's add_camera() method to create all camera
        interface instances.

    Args:
        system_id: The unique identifier code of the VideoSystem instance that uses this camera interface.
        frame_rate: The simulated frame acquisition rate of the camera, in frames per second.
        frame_width: The simulated camera frame width, in pixels.
        frame_height: The simulated camera frame height, in pixels.
        color: The simulated camera frame color mode. If True, the frames are generated using the BGR color mode. If
            False, the frames are generated using the grayscale (monochrome) color mode.

    Attributes:
        _system_id: Stores the unique identifier code of the VideoSystem instance that uses this camera interface.
        _color: Determines whether to simulate monochrome or RGB frame images.
        _camera: Tracks whether the camera is 'connected'.
        _frame_rate: Stores the camera's frame acquisition rate.
        _frame_width: Stores the width of the camera's frames.
        _frame_height: Stores the height of the camera's frames.
        _acquiring: Tracks whether the camera is currently acquiring video frames.
        _frames: Stores the pool of pre-generated frame images used to simulate camera frame acquisition.
        _current_frame_index: The index of the currently evaluated frame in the pre-generated frame pool buffer. This
            is used to simulate the behavior of the cyclic buffer used by physical cameras.
        _timer: After the camera is 'connected', this attribute is used to store the timer class that controls the
            simulated camera's frame rate.
        _time_between_frames: Stores the number of milliseconds that has to pass between two consecutive frame
            acquisitions, used to simulate a physical camera's frame rate.
    """

    def __init__(
        self,
        system_id: int,
        frame_rate: int | None = None,
        frame_width: int | None = None,
        frame_height: int | None = None,
        *,
        color: bool = True,
    ) -> None:
        # Saves class parameters to class attributes
        self._system_id: int = system_id
        self._color: bool = color
        self._frame_rate: int = 30 if frame_rate is None else frame_rate
        self._frame_width: int = 600 if frame_width is None else frame_width
        self._frame_height: int = 400 if frame_height is None else frame_height
        self._camera: bool = False
        self._acquiring: bool = False

        # Creates a random number generator to be sued below
        rng = np.random.default_rng(seed=42)  # Specifies a reproducible seed.

        # To allow reproducible testing, the class statically generates a pool of 10 images used during the grab_frame()
        # method calls.
        frames_list: list[NDArray[np.uint8]] = []
        for _ in range(10):
            if self._color:
                frame = rng.integers(0, 256, size=(self._frame_height, self._frame_width, 3), dtype=np.uint8)
                # Ensures the order of the colors is BGR
                bgr_frame: NDArray[np.uint8] = cv2.cvtColor(  # type: ignore[assignment]
                    src=frame, code=cv2.COLOR_RGB2BGR
                )
                frames_list.append(bgr_frame)
            else:
                # grayscale frames have only one channel, so order does not matter
                frames_list.append(
                    rng.integers(0, 256, size=(self._frame_height, self._frame_width, 1), dtype=np.uint8)
                )

        # Casts to a tuple to optimize runtime efficiency
        self._frames: tuple[NDArray[np.uint8], ...] = tuple(frames_list)
        self._current_frame_index: int = 0

        # Cannot be initialized here due to the use of multiprocessing in the VideoSystem class.
        self._timer: PrecisionTimer | None = None

        # Uses the frame_rate to derive the number of microseconds that has to pass between each frame acquisition.
        # This is used to simulate the camera's frame rate during grab_frame() runtime.
        self._time_between_frames: float = 1000 / self._frame_rate

    def connect(self) -> None:
        """Simulates connecting to the camera hardware."""
        self._camera = True

        # Uses millisecond precision, which supports simulating up to 1000 fps. The time has to be initialized here to
        # make the class compatible with the VideoSystem class that uses multiprocessing.
        self._timer = PrecisionTimer("ms")

    def disconnect(self) -> None:
        """Simulates disconnecting from the camera hardware."""
        self._camera = False
        self._acquiring = False
        self._timer = None

    @property
    def is_connected(self) -> bool:
        """Returns True if the instance is 'connected' to the camera hardware."""
        return self._camera

    @property
    def is_acquiring(self) -> bool:
        """Returns True if the camera is currently 'acquiring' video frames."""
        return self._acquiring

    @property
    def frame_rate(self) -> int:
        """Returns the acquisition rate of the camera, in frames per second (fps)."""
        return self._frame_rate

    @property
    def frame_width(self) -> int:
        """Returns the width of the acquired frames, in pixels."""
        return self._frame_width

    @property
    def frame_height(self) -> int:
        """Returns the height of the acquired frames, in pixels."""
        return self._frame_height

    @property
    def frame_pool(self) -> tuple[NDArray[np.uint8], ...]:
        """Returns the pool of camera frames sampled by the grab_frame() method."""
        return self._frames

    @property
    def pixel_color_format(self) -> InputPixelFormats:
        """Returns the pixel color format of the acquired frames."""
        if self._color:
            return InputPixelFormats.BGR
        return InputPixelFormats.MONOCHROME

    def grab_frame(self) -> NDArray[np.uint8]:
        """Grabs the first available frame from the managed camera's acquisition buffer.

        This method has to be called repeatedly (cyclically) to fetch the newly acquired frames from the camera.

        Returns:
            A NumPy array that stores the frame data. Depending on whether the camera acquires colored or monochrome
            images, the returned arrays have the shape (height, width, channels) or (height, width). Color data uses
            the BGR channel order.

        Raises:
            RuntimeError: If the method is called for a class not currently 'connected' to a camera.
        """
        # Prevents calling this method before connecting to the camera's hardware
        if not self._camera:
            message = (
                f"The MockCamera instance for the VideoSystem with id {self._system_id} is not currently simulating "
                f"connection to the camera hardware, and cannot simulate image acquisition. Call the connect() method "
                f"prior to calling the grab_frame() method."
            )
            console.error(message=message, error=ConnectionError)
            # Fallback to appease mypy, should not be reachable
            raise ConnectionError(message)  # pragma: no cover

        # Flips the acquiring flag the first time this method is called
        if not self._acquiring:
            self._acquiring = True

        # Fallback to appease mypy, the time should always be initialized at this point
        if self._timer is None:
            self._timer = PrecisionTimer("ms")

        # All camera interfaces are designed to block in-place if the frame is not available. Here, this behavior
        # is simulated by using the timer class to 'force' the method to work at a certain frame rate.
        while self._timer.elapsed < self._time_between_frames:
            pass

        # 'Acquires' a frame from the frame pool
        frame = self._frames[self._current_frame_index].copy()

        # Resets the timer to measure the time elapsed since the last frame acquisition.
        self._timer.reset()

        # Increments the flame pool index. When the index reaches the end of the pool, this resets it back to the
        # starts of the popol. This simulates the behavior of a cyclic buffer.
        self._current_frame_index = (self._current_frame_index + 1) % _FRAME_POOL_SIZE

        # Returns the acquired frame to caller
        return frame



"""Contains tests for classes and methods provided by the video_system.py module."""

import sys
import subprocess

import numpy as np
import pytest
from ataraxis_time import PrecisionTimer
from ataraxis_base_utilities import error_format
from ataraxis_data_structures import DataLogger, assemble_log_archives

from ataraxis_video_system import VideoSystem
from ataraxis_video_system.saver import (
    VideoEncoders,
    OutputPixelFormats,
    EncoderSpeedPresets,
    check_ffmpeg_availability,
)
from ataraxis_video_system.camera import CameraInterfaces, get_harvesters_ids, get_opencv_ids
from ataraxis_video_system.video_system import extract_logged_camera_timestamps
from random import randint


@pytest.fixture(scope="session")
def has_opencv():
    """Checks for OpenCV camera availability in the test environment."""
    try:
        opencv_ids = get_opencv_ids()
        if len(opencv_ids) > 0:
            return True
        else:
            return False
    except:
        return False


@pytest.fixture(scope="session")
def has_harvesters():
    """Checks for Harvesters camera availability in the test environment."""
    try:
        # Attempts to discover Harvesters cameras using the internally stored CTI path
        harvesters_ids = get_harvesters_ids()
        if len(harvesters_ids) > 0:
            return True
        else:
            return False
    except:
        return False


@pytest.fixture(scope="session")
def has_nvidia():
    """Checks for NVIDIA GPU availability in the test environment."""
    try:
        subprocess.run(
            ["nvidia-smi", "--query-gpu=name", "--format=csv,noheader"],
            capture_output=True,
            text=True,
            check=True,
            timeout=30,
        )
        return True
    except:
        return False


@pytest.fixture
def data_logger(tmp_path) -> DataLogger:
    """Creates a DataLogger instance and returns it to the caller."""
    data_logger = DataLogger(output_directory=tmp_path, instance_name=str(randint(0, 100000000000)))
    return data_logger


@pytest.fixture
def video_system(tmp_path, data_logger) -> VideoSystem:
    """Creates a VideoSystem instance and returns it to the caller."""
    system_id = np.uint8(1)
    output_directory = tmp_path.joinpath("test_output_directory")

    return VideoSystem(
        system_id=system_id,
        data_logger=data_logger,
        output_directory=output_directory,
        camera_interface=CameraInterfaces.MOCK,
        camera_index=0,
        display_frame_rate=None,
        frame_width=None,
        frame_height=None,
        frame_rate=None,
        gpu=-1,
        video_encoder=VideoEncoders.H265,
        encoder_speed_preset=EncoderSpeedPresets.SLOW,
        output_pixel_format=OutputPixelFormats.YUV444,
        quantization_parameter=15,
        color=True,
    )


def test_init_repr(tmp_path, data_logger) -> None:
    """Verifies the functioning of the VideoSystem __init__() and __repr__() methods."""
    vs_instance = VideoSystem(
        system_id=np.uint8(1),
        data_logger=data_logger,
        output_directory=tmp_path.joinpath("test_output_directory"),
        camera_interface=CameraInterfaces.MOCK,
        camera_index=0,
    )

    # Verifies class properties
    assert vs_instance.system_id == np.uint8(1)
    assert not vs_instance.started
    assert vs_instance.video_file_path is not None

    # Verifies the __repr()__ method
    representation_string: str = (
        f"VideoSystem(system_id={np.uint8(1)}, started={False}, camera=MockCamera, frame_saving={True})"
    )
    assert repr(vs_instance) == representation_string


def test_init_errors(data_logger) -> None:
    """Verifies the error handling behavior of the VideoSystem initialization method."""
    # Invalid system_id input - causes conversion error
    invalid_system_id = "str"
    # noinspection PyTypeChecker
    with pytest.raises((TypeError, ValueError)):
        VideoSystem(
            system_id=invalid_system_id,  # type: ignore
            data_logger=data_logger,
            output_directory=data_logger.output_directory,
            camera_interface=CameraInterfaces.MOCK,
        )

    # Invalid data_logger input
    invalid_data_logger = None
    message = (
        f"Unable to initialize the VideoSystem instance with id 1. Expected an initialized "
        f"DataLogger instance as the 'data_logger' argument value, but encountered {invalid_data_logger} of type "
        f"{type(invalid_data_logger).__name__}."
    )
    with pytest.raises(TypeError, match=error_format(message)):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=invalid_data_logger,  # type: ignore
            output_directory=data_logger.output_directory,
        )

    # Invalid output_directory input
    invalid_output_directory = "Not a Path"
    message = (
        f"Unable to initialize the VideoSystem instance with id 1. Expected a Path instance or None "
        f"as the 'output_directory' argument's value, but encountered {invalid_output_directory} of type "
        f"{type(invalid_output_directory).__name__}."
    )
    with pytest.raises(TypeError, match=error_format(message)):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=invalid_output_directory,  # type: ignore
            camera_interface=CameraInterfaces.MOCK,
        )


def test_camera_configuration_errors(data_logger, tmp_path) -> None:
    """Verifies the error handling behavior of camera configuration during VideoSystem initialization."""
    output_directory = tmp_path.joinpath("test_output_directory")

    # Invalid camera index
    invalid_index = "str"
    message = (
        f"Unable to configure the camera interface for the VideoSystem with id 1. Expected a "
        f"zero or positive integer as the 'camera_id' argument value, but got {invalid_index} of type "
        f"{type(invalid_index).__name__}."
    )
    with pytest.raises(TypeError, match=error_format(message)):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            camera_index=invalid_index,  # type: ignore
            camera_interface=CameraInterfaces.MOCK,
        )

    # Invalid frame rate
    invalid_frame_rate = "str"
    message = (
        f"Unable to configure the camera interface for the VideoSystem with id 1. Expected a "
        f"positive integer or None as the 'frame_rate' argument value, but got "
        f"{invalid_frame_rate} of type {type(invalid_frame_rate).__name__}."
    )
    with pytest.raises(TypeError, match=error_format(message)):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            frame_rate=invalid_frame_rate,  # type: ignore
            camera_interface=CameraInterfaces.MOCK,
        )

    # Invalid frame width
    invalid_frame_width = "str"
    message = (
        f"Unable to configure the camera interface for the VideoSystem with id 1. Expected a "
        f"positive integer or None as the 'frame_width' argument value, but got {invalid_frame_width} of type "
        f"{type(invalid_frame_width).__name__}."
    )
    with pytest.raises(TypeError, match=error_format(message)):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            frame_width=invalid_frame_width,  # type: ignore
            camera_interface=CameraInterfaces.MOCK,
        )

    # Invalid frame height
    invalid_frame_height = "str"
    message = (
        f"Unable to configure the camera interface for the VideoSystem with id 1. Expected a "
        f"positive integer or None as the 'frame_height' argument value, but got {invalid_frame_height} of type "
        f"{type(invalid_frame_height).__name__}."
    )
    with pytest.raises(TypeError, match=error_format(message)):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            frame_height=invalid_frame_height,  # type: ignore
            camera_interface=CameraInterfaces.MOCK,
        )

    # Invalid camera interface
    invalid_interface = "invalid"
    message_pattern = "Unable to configure the camera interface.*unsupported camera_interface"
    with pytest.raises(ValueError, match=message_pattern):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            camera_interface=invalid_interface,  # type: ignore
        )


def test_video_saver_configuration(data_logger, tmp_path, has_nvidia) -> None:
    """Verifies the functioning of video saver configuration during VideoSystem initialization."""
    output_directory = tmp_path.joinpath("test_output_directory")

    # Tests GPU encoding if NVIDIA GPU is available, otherwise tests CPU encoding
    if has_nvidia and check_ffmpeg_availability():
        vs = VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            camera_interface=CameraInterfaces.MOCK,
            gpu=0,
            video_encoder=VideoEncoders.H265,
            encoder_speed_preset=EncoderSpeedPresets.FASTEST,
            output_pixel_format=OutputPixelFormats.YUV444,
            quantization_parameter=5,
        )
        assert vs._saver is not None
    elif check_ffmpeg_availability():
        vs = VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            camera_interface=CameraInterfaces.MOCK,
            gpu=-1,
            video_encoder=VideoEncoders.H265,
            encoder_speed_preset=EncoderSpeedPresets.FASTEST,
            output_pixel_format=OutputPixelFormats.YUV444,
            quantization_parameter=5,
        )
        assert vs._saver is not None


def test_video_saver_configuration_errors(data_logger, tmp_path) -> None:
    """Verifies the error handling behavior of video saver configuration during VideoSystem initialization."""
    output_directory = tmp_path.joinpath("test_output_directory")

    # Invalid gpu index
    invalid_gpu = "str"
    message = (
        f"Unable to configure the video saver for the VideoSystem with id 1. Expected an "
        f"integer as the 'gpu' argument value, but got {invalid_gpu} of type {type(invalid_gpu).__name__}."
    )
    with pytest.raises(TypeError, match=error_format(message)):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            gpu=invalid_gpu,  # type: ignore
            camera_interface=CameraInterfaces.MOCK,
        )

    # Invalid video encoder
    invalid_encoder = "invalid"
    message_pattern = "Unable to configure the video saver.*unexpected 'video_encoder'"
    with pytest.raises(ValueError, match=message_pattern):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            video_encoder=invalid_encoder,  # type: ignore
            camera_interface=CameraInterfaces.MOCK,
        )

    # Invalid encoder preset
    invalid_preset = "invalid"
    message_pattern = "Unable to configure the video saver.*unexpected 'encoder_speed_preset'"
    with pytest.raises(ValueError, match=message_pattern):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            encoder_speed_preset=invalid_preset,  # type: ignore
            camera_interface=CameraInterfaces.MOCK,
        )

    # Invalid output pixel format
    invalid_format = "invalid"
    message_pattern = "Unable to configure the video saver.*unexpected 'output_pixel_format'"
    with pytest.raises(ValueError, match=message_pattern):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            output_pixel_format=invalid_format,  # type: ignore
            camera_interface=CameraInterfaces.MOCK,
        )

    # Invalid quantization parameter
    invalid_qp = "str"
    message = (
        f"Unable to configure the video saver for the VideoSystem with id 1. Expected an "
        f"integer between -1 and 51 as the 'quantization_parameter' argument value, but got "
        f"{invalid_qp} of type {type(invalid_qp).__name__}."
    )
    with pytest.raises(TypeError, match=error_format(message)):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            quantization_parameter=invalid_qp,  # type: ignore
            camera_interface=CameraInterfaces.MOCK,
        )


def test_start_stop(data_logger, tmp_path) -> None:
    """Verifies the functioning of the start(), stop(), start_frame_saving() and stop_frame_saving() methods of the
    VideoSystem class.

    Also verifies internal DataLogger bindings and logged timestamp extraction methods.
    """
    # Creates two VideoSystem instances with different configurations
    output_directory = tmp_path.joinpath("test_output_directory")

    video_system_1 = VideoSystem(
        system_id=np.uint8(101),
        data_logger=data_logger,
        output_directory=output_directory,
        camera_interface=CameraInterfaces.MOCK,
        frame_rate=10,
        display_frame_rate=None,
        quantization_parameter=40,
    )

    video_system_2 = VideoSystem(
        system_id=np.uint8(202),
        data_logger=data_logger,
        output_directory=None,  # No saving configured
        camera_interface=CameraInterfaces.MOCK,
        frame_rate=5,
    )

    # Starts all instances
    data_logger.start()
    video_system_1.start()
    video_system_1.start()  # Ensures that calling start twice does nothing
    video_system_2.start()

    assert video_system_1.started
    assert video_system_2.started

    # Tests frame saving control
    timer = PrecisionTimer("s")
    video_system_1.start_frame_saving()
    timer.delay(delay=2, allow_sleep=True, block=False)  # 2-second delay
    video_system_1.stop_frame_saving()

    # Tests that system 2 (without a saver) ignores saving commands
    video_system_2.start_frame_saving()
    video_system_2.stop_frame_saving()

    # Re-enables saving for system 1
    video_system_1.start_frame_saving()
    timer.delay(delay=2, allow_sleep=True, block=False)  # 2-second delay
    video_system_1.stop_frame_saving()

    # Stops the video systems
    video_system_1.stop()
    video_system_2.stop()
    video_system_2.stop()  # Ensures that calling stop twice does nothing

    assert not video_system_1.started
    assert not video_system_2.started

    # Compresses logs for timestamp extraction
    assemble_log_archives(log_directory=data_logger.output_directory, remove_sources=True, memory_mapping=False)

    # Extracts frame timestamps for system 1 (which saved frames)
    log_path_1 = data_logger.output_directory.joinpath(f"{data_logger._name}_101.npz")
    if log_path_1.exists():
        frame_timestamps_1 = extract_logged_camera_timestamps(log_path_1, n_workers=1)
        # With fps of 10 and running for ~4 seconds total, should have acquired around 40 frames
        assert 35 <= len(frame_timestamps_1) <= 45

    # Tests the system without frame saving
    video_system_3 = VideoSystem(
        system_id=np.uint8(234),
        data_logger=data_logger,
        output_directory=None,  # No output directory
        camera_interface=CameraInterfaces.MOCK,
    )
    video_system_3.start()
    timer.delay(delay=1, allow_sleep=True, block=False)
    video_system_3.stop()
    data_logger.stop()


def test_display_frame_rate_validation(data_logger, tmp_path) -> None:
    """Verifies the validation of the display_frame_rate parameter."""
    output_directory = tmp_path.joinpath("test_output_directory")

    # Tests that display functionality is disabled on macOS
    if "darwin" in sys.platform:
        with pytest.warns(UserWarning, match="Displaying frames is currently not supported"):
            vs = VideoSystem(
                system_id=np.uint8(1),
                data_logger=data_logger,
                output_directory=output_directory,
                camera_interface=CameraInterfaces.MOCK,
                display_frame_rate=30,
            )
            assert vs._display_frame_rate == 0

    # Tests invalid display frame rate (string instead of int)
    invalid_display_rate = "str"
    message = (
        f"Unable to configure the camera interface for the VideoSystem with id 1. Encountered "
        f"an unsupported 'display_frame_rate' argument value {invalid_display_rate} of type "
        f"{type(invalid_display_rate).__name__}. The display frame rate override has to be None or a positive "
        f"integer that does not exceed the camera acquisition frame rate (30)."
    )
    with pytest.raises(TypeError, match=error_format(message)):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            camera_interface=CameraInterfaces.MOCK,
            frame_rate=30,
            display_frame_rate=invalid_display_rate,  # type: ignore
        )

    # Tests display frame rate exceeding acquisition rate
    excessive_display_rate = 60
    message = (
        f"Unable to configure the camera interface for the VideoSystem with id 1. Encountered "
        f"an unsupported 'display_frame_rate' argument value {excessive_display_rate} of type "
        f"{type(excessive_display_rate).__name__}. The display frame rate override has to be None or a positive "
        f"integer that does not exceed the camera acquisition frame rate (30)."
    )
    with pytest.raises(TypeError, match=error_format(message)):
        VideoSystem(
            system_id=np.uint8(1),
            data_logger=data_logger,
            output_directory=output_directory,
            camera_interface=CameraInterfaces.MOCK,
            frame_rate=30,
            display_frame_rate=excessive_display_rate,  # Exceeds frame_rate
        )


def test_extract_logged_camera_timestamps_errors(tmp_path) -> None:
    """Verifies the error handling of the extract_logged_camera_timestamps() function."""
    # Tests with a non-existent file
    non_existent_path = tmp_path.joinpath("non_existent.npz")
    message = (
        f"Unable to extract camera frame timestamp data from the log file {non_existent_path}, as it does not exist "
        f"or does not point to a valid .npz archive."
    )
    with pytest.raises(ValueError, match=error_format(message)):
        extract_logged_camera_timestamps(non_existent_path)

    # Tests with invalid file extension
    invalid_path = tmp_path.joinpath("invalid.txt")
    invalid_path.touch()
    message = (
        f"Unable to extract camera frame timestamp data from the log file {invalid_path}, as it does not exist "
        f"or does not point to a valid .npz archive."
    )
    with pytest.raises(ValueError, match=error_format(message)):
        extract_logged_camera_timestamps(invalid_path)


def test_camera_timestamp_extraction(data_logger, tmp_path) -> None:
    """Tests timestamp extraction with start/stop segments of frame saving.

    This test verifies that timestamps are correctly extracted when frame saving
    is enabled and disabled multiple times during a single session.
    """

    system_id = np.uint8(99)
    frame_rate = 10  # Lower frame rate for easier validation

    output_directory = tmp_path.joinpath("test_segmented_timestamps")
    output_directory.mkdir(parents=True, exist_ok=True)

    video_system = VideoSystem(
        system_id=system_id,
        data_logger=data_logger,
        output_directory=output_directory,
        camera_interface=CameraInterfaces.MOCK,
        frame_rate=frame_rate,
        frame_width=320,
        frame_height=240,
        color=True,
    )

    # Start systems
    data_logger.start()
    video_system.start()

    timer = PrecisionTimer("s")

    # First segment: 1 second of recording
    video_system.start_frame_saving()
    timer.delay(delay=1, allow_sleep=True, block=False)
    video_system.stop_frame_saving()

    # Pause: 1 second without recording
    timer.delay(delay=1, allow_sleep=True, block=False)

    # Second segment: 2 seconds of recording
    video_system.start_frame_saving()
    timer.delay(delay=2, allow_sleep=True, block=False)
    video_system.stop_frame_saving()

    # Pause: 1 second without recording
    timer.delay(delay=1, allow_sleep=True, block=False)

    # Third segment: 1 second of recording
    video_system.start_frame_saving()
    timer.delay(delay=1, allow_sleep=True, block=False)
    video_system.stop_frame_saving()

    # Stop systems
    video_system.stop()
    data_logger.stop()

    # Process logs
    assemble_log_archives(log_directory=data_logger.output_directory, remove_sources=True, memory_mapping=False)

    # Extracts timestamps
    log_file_path = data_logger.output_directory.joinpath(f"{system_id}_log.npz")
    timestamps = extract_logged_camera_timestamps(log_file_path, n_workers=1)

    # Total recording time: 1 + 2 + 1 = 4 seconds
    # Expected frames: approximately 40 (4 * 10 fps)
    actual_frames = len(timestamps)

    # Allow for timing variations
    assert 0 <= actual_frames, f"Expected approximately {40} frames, got {actual_frames}"

    # Check for gaps in timestamps that might indicate the pauses
    # (This is a basic check - actual gaps depend on implementation details)
    if len(timestamps) > 10:
        intervals = [timestamps[i] - timestamps[i - 1] for i in range(1, len(timestamps))]
        max_interval = max(intervals)
        avg_interval = np.mean(intervals)

        # The maximum interval might be larger due to pauses, but shouldn't be
        # excessive (e.g., not more than 10x the average for this controlled test)
        assert max_interval < avg_interval * 10, "Detected unexpectedly large gap in timestamps"



"""A Python library that interfaces with a wide range of cameras to flexibly record visual stream data as video files.

See https://github.com/Sun-Lab-NBB/ataraxis-video-system for more details.
API documentation: https://ataraxis-video-system-api-docs.netlify.app/
Authors: Ivan Kondratyev (Inkaros), Jacob Groner, Natalie Yeung
"""

import os
import multiprocessing as mp

# Applies important library-wide configurations to optimize runtime performance.
if mp.get_start_method(allow_none=True) is None:
    # Makes the library behave the same way across all platforms.
    mp.set_start_method("spawn")

# Improves frame rendering (display) on Windows operating systems.
os.environ["OPENCV_VIDEOIO_MSMF_ENABLE_HW_TRANSFORMS"] = "0"

# The QT bundled with OpenCV (used for live image rendering) does not include the wayland support plugin. This forces
# QT to use the X11 compatibility layer when it is called from a Wayland system.
if "WAYLAND_DISPLAY" in os.environ:
    os.environ["QT_QPA_PLATFORM"] = "xcb"


from .saver import (
    VideoEncoders,
    InputPixelFormats,
    OutputPixelFormats,
    EncoderSpeedPresets,
    check_gpu_availability,
    check_ffmpeg_availability,
)
from .camera import CameraInterfaces, CameraInformation, add_cti_file, get_opencv_ids, get_harvesters_ids
from .video_system import VideoSystem, extract_logged_camera_timestamps

__all__ = [
    "CameraInformation",
    "CameraInterfaces",
    "EncoderSpeedPresets",
    "InputPixelFormats",
    "OutputPixelFormats",
    "VideoEncoders",
    "VideoSystem",
    "add_cti_file",
    "check_ffmpeg_availability",
    "check_gpu_availability",
    "extract_logged_camera_timestamps",
    "get_harvesters_ids",
    "get_opencv_ids",
]



"""This module provides the main VideoSystem class that contains methods for setting up, running, and tearing down
interactions between camera interfaces and video saver instances. All user-oriented functionality of this library is
available through the public methods of the VideoSystem class.
"""

import sys
from queue import Empty, Queue
from typing import TYPE_CHECKING, Any
from pathlib import Path
import warnings
from threading import Thread
from multiprocessing import (
    Queue as MPQueue,
    Manager,
    Process,
    cpu_count,
)
from concurrent.futures import ProcessPoolExecutor, as_completed

import cv2
from tqdm import tqdm
import numpy as np
from ataraxis_time import PrecisionTimer
from ataraxis_base_utilities import console, chunk_iterable
from ataraxis_data_structures import DataLogger, LogPackage, SharedMemoryArray
from ataraxis_time.time_helpers import TimestampFormats, convert_time, get_timestamp

from .saver import (
    VideoSaver,
    VideoEncoders,
    OutputPixelFormats,
    EncoderSpeedPresets,
    check_gpu_availability,
    check_ffmpeg_availability,
)
from .camera import MockCamera, OpenCVCamera, CameraInterfaces, HarvestersCamera

if TYPE_CHECKING:
    from multiprocessing.managers import SyncManager

    from numpy.typing import NDArray


# Determines the maximum qp value used when initializing VideoSystem instances.
_MAXIMUM_QUANTIZATION_VALUE = 51

# The maximum number of seconds to wait for the consumer and producer processes to initialize.
_PROCESS_INITIALIZATION_TIME = 20

# The maximum number of seconds to wait for the consumer and producer processes to shut down.
_PROCESS_SHUTDOWN_TIME = 600

# Specifies the size of the frame timestamp log message, in bytes. This is used during log message parsing to extract
# the acquisition timestamps for all saved frames.
_FRAME_TIMESTAMP_LOG_MESSAGE_SIZE = 9

# The minimum number of messages that must be contained in the camera frame timestamp .npz archive for it to be
# processed in parallel.
_MINIMUM_LOG_SIZE_FOR_PARALLELIZATION = 2000


class VideoSystem:
    """Acquires, displays, and saves camera frames to disk using the requested camera interface and video saver.

    This class controls the runtime of a camera interface and a video saver running in independent processes and
    efficiently moves the frames acquired by the camera to the saver process.

    Notes:
        This class reserves up to two logical cores to support the producer (camera interface) and consumer
        (video saver) processes. Additionally, it reserves a variable portion of the RAM to buffer the frames as they
        are moved from the producer to the consumer.

        Video saving relies on the third-party software 'FFMPEG' to encode the video frames as an .mp4 file.
        See https://www.ffmpeg.org/download.html for more information on installing the library.

    Args:
        system_id: The unique value to use for identifying the VideoSystem instance in all output streams (log files,
            terminal messages, video files).
        data_logger: An initialized DataLogger instance used to log the timestamps for all frames saved by this
            VideoSystem instance.
        output_directory: The path to the output directory where to store the acquired frames as the .mp4 video file.
            Setting this argument to None disabled video saving functionality.
        camera_interface: The interface to use for working with the camera hardware. Must be one of the CameraInterfaces
            enumeration members.
        camera_index: The index of the camera in the list of all cameras discoverable by the chosen interface, e.g.: 0
            for the first available camera, 1 for the second, etc. This specifies the camera hardware the instance
            should interface with at runtime.
        display_frame_rate: Determines the frame rate at which to display the acquired frames to the user. Setting this
            argument to None (default) disables frame display functionality. Note, frame displaying is not supported
            on some macOS versions.
        frame_rate: The desired rate, in frames per second, at which to capture the frames. Note; whether the requested
            rate is attainable depends on the hardware capabilities of the camera and the communication interface. If
            this argument is not explicitly provided, the instance uses the default frame rate of the managed camera.
        frame_width: The desired width of the acquired frames, in pixels. Note; the requested width must be compatible
            with the range of frame dimensions supported by the camera hardware. If this argument is not explicitly
            provided, the instance uses the default frame width of the managed camera.
        frame_height: Same as 'frame_width', but specifies the desired height of the acquired frames, in pixels. If this
            argument is not explicitly provided, the instance uses the default frame width of the managed camera.
        color: Specifies whether the camera acquires colored or monochrome images. This determines how to store the
            acquired frames. Colored frames are saved using the 'BGR' channel order, monochrome images are reduced to
            a single-channel format. This argument is only used by the OpenCV and Mock camera interfaces, the
            Harvesters interface infers this information directly from the camera's configuration.
        gpu: The index of the GPU to use for video encoding. Setting this argument to a value of -1 (default) configures
            the instance to use the CPU for encoding. Valid GPU indices can be obtained from the 'nvidia-smi' terminal
            command.
        video_encoder: The encoder to use for generating the video file. Must be one of the valid VideoEncoders
            enumeration members.
        encoder_speed_preset: The encoding speed preset to use for generating the video file. Must be one of the valid
            EncoderSpeedPresets enumeration members.
        output_pixel_format: The pixel format to be used by the output video file. Must be one of the valid
            OutputPixelFormats enumeration members.
        quantization_parameter: The integer value to use for the 'quantization parameter' of the encoder. This
            determines how much information to discard from each encoded frame. Lower values produce better video
            quality at the expense of longer processing time and larger file size: 0 is best, 51 is worst. Note, the
            default value is calibrated for the H265 encoder and is likely too low for the H264 encoder.

    Attributes:
        _started: Tracks whether the system is currently running (has active subprocesses).
        _mp_manager: Stores the SyncManager instance used to control the multiprocessing assets (Queue and Lock
            instances).
        _system_id: Stores the unique identifier code of the VideoSystem instance.
        _output_file: Stores the path to the output .mp4 video file to be generated at runtime or None, if the instance
            is not configured to save acquired camera frames.
        _camera: Stores the camera interface class instance used to interface with the camera hardware at runtime.
        _saver: Stores the video saver instance used to save the acquired camera frames or None, if the instance is
            not configured to save acquired camera frames.
        _logger_queue: Stores the multiprocessing Queue instance used to buffer frame acquisition timestamp data to the
            logger process.
        _saver_queue: Stores the multiprocessing Queue instance used to buffer and pipe acquired frames from the
            camera (producer) process to the video saver (consumer) process.
        _terminator_array: Stores the SharedMemoryArray instance used to manage the runtime behavior of the producer
            and consumer processes.
        _producer_process: A process that acquires camera frames using the managed camera interface.
        _consumer_process: A process that saves the acquired frames using managed video saver.
        _watchdog_thread: A thread used to monitor the runtime status of the remote consumer and producer processes.

    Raises:
        TypeError: If any of the provided arguments has an invalid type.
        ValueError: If any of the provided arguments has an invalid value.
        RuntimeError: If the host system does not have access to FFMPEG or Nvidia GPU (when the instance is configured
            to use hardware encoding).
    """

    def __init__(
        self,
        system_id: np.uint8,
        data_logger: DataLogger,
        output_directory: Path | None,
        camera_interface: CameraInterfaces | str = CameraInterfaces.OPENCV,
        camera_index: int = 0,
        display_frame_rate: int | None = None,
        frame_width: int | None = None,
        frame_height: int | None = None,
        frame_rate: int | None = None,
        gpu: int = -1,
        video_encoder: VideoEncoders | str = VideoEncoders.H265,
        encoder_speed_preset: EncoderSpeedPresets | int = EncoderSpeedPresets.SLOW,
        output_pixel_format: OutputPixelFormats | str = OutputPixelFormats.YUV444,
        quantization_parameter: int = 15,
        *,
        color: bool | None = None,
    ) -> None:
        # Has to be set first to avoid stop method errors
        self._started: bool = False  # Tracks whether the system has active processes

        # The manager is created early in the __init__ phase to support del-based cleanup
        self._mp_manager: SyncManager = Manager()

        # Ensures system_id is a byte-convertible integer
        self._system_id: np.uint8 = np.uint8(system_id)

        # Ensures that the data_logger is an initialized DataLogger instance.
        if not isinstance(data_logger, DataLogger):
            message = (
                f"Unable to initialize the VideoSystem instance with id {system_id}. Expected an initialized "
                f"DataLogger instance as the 'data_logger' argument value, but encountered {data_logger} of type "
                f"{type(data_logger).__name__}."
            )
            console.error(message=message, error=TypeError)

        # Ensures that the output_directory is either a Path instance or None:
        if output_directory is not None and not isinstance(output_directory, Path):
            message = (
                f"Unable to initialize the VideoSystem instance with id {system_id}. Expected a Path instance or None "
                f"as the 'output_directory' argument's value, but encountered {output_directory} of type "
                f"{type(output_directory).__name__}."
            )
            console.error(message=message, error=TypeError)

        # If the output directory is provided, resolves the path to the output .mp4 video file to be created during
        # runtime.
        self._output_file: Path | None = (
            None if output_directory is None else output_directory.joinpath(f"{system_id:03d}.mp4")
        )

        # Initializes the camera interface:

        # Validates camera-related inputs:
        if not isinstance(camera_index, int) or camera_index < 0:
            message = (
                f"Unable to configure the camera interface for the VideoSystem with id {self._system_id}. Expected a "
                f"zero or positive integer as the 'camera_id' argument value, but got {camera_index} of type "
                f"{type(camera_index).__name__}."
            )
            console.error(error=TypeError, message=message)
        if (frame_rate is not None and not isinstance(frame_rate, int)) or (
            isinstance(frame_rate, int) and frame_rate <= 0
        ):
            message = (
                f"Unable to configure the camera interface for the VideoSystem with id {self._system_id}. Expected a "
                f"positive integer or None as the 'frame_rate' argument value, but got "
                f"{frame_rate} of type {type(frame_rate).__name__}."
            )
            console.error(error=TypeError, message=message)
        if (frame_width is not None and not isinstance(frame_width, int)) or (
            isinstance(frame_width, int) and frame_width <= 0
        ):
            message = (
                f"Unable to configure the camera interface for the VideoSystem with id {self._system_id}. Expected a "
                f"positive integer or None as the 'frame_width' argument value, but got {frame_width} of type "
                f"{type(frame_width).__name__}."
            )
            console.error(error=TypeError, message=message)
        if (frame_height is not None and not isinstance(frame_height, int)) or (
            isinstance(frame_height, int) and frame_height <= 0
        ):
            message = (
                f"Unable to configure the camera interface for the VideoSystem with id {self._system_id}. Expected a "
                f"positive integer or None as the 'frame_height' argument value, but got {frame_height} of type "
                f"{type(frame_height).__name__}."
            )
            console.error(error=TypeError, message=message)

        # Presets the variable type
        self._camera: OpenCVCamera | HarvestersCamera | MockCamera

        # OpenCVCamera
        if camera_interface == CameraInterfaces.OPENCV:
            # Instantiates the OpenCVCamera object
            self._camera = OpenCVCamera(
                system_id=int(self._system_id),
                color=False if not isinstance(color, bool) else color,
                camera_index=camera_index,
                frame_height=frame_height,
                frame_width=frame_width,
                frame_rate=frame_rate,
            )

        # HarvestersCamera
        elif camera_interface == CameraInterfaces.HARVESTERS:
            # Instantiates the HarvestersCamera object
            self._camera = HarvestersCamera(
                system_id=int(self._system_id),
                camera_index=camera_index,
                frame_height=frame_height,
                frame_width=frame_width,
                frame_rate=frame_rate,
            )

        # MockCamera
        elif camera_interface == CameraInterfaces.MOCK:
            # Instantiates the MockCamera object
            self._camera = MockCamera(
                system_id=int(self._system_id),
                frame_height=frame_height,
                frame_width=frame_width,
                frame_rate=frame_rate,
                color=False if not isinstance(color, bool) else color,
            )

        # If the requested camera interface does not match any of the supported interfaces, raises an error
        else:
            message = (
                f"Unable to configure the camera interface for the VideoSystem with id {self._system_id}. Encountered "
                f"an unsupported camera_interface argument value {camera_interface} of type "
                f"{type(camera_interface).__name__}. Use one of the supported CameraInterfaces enumeration members: "
                f"{', '.join(tuple(camera_interface))}."
            )
            console.error(error=ValueError, message=message)
            # Fallback to appease mypy, should not be reachable
            raise ValueError(message)  # pragma: no cover

        # Connects to the camera. This both verifies that the camera can be connected to and applies the camera
        # acquisition parameters.
        self._camera.connect()

        # Verifies that the frame acquisition works as expected.
        self._camera.grab_frame()

        # Disconnects from the camera. The camera is re-connected by the remote producer process once it is
        # instantiated.
        self._camera.disconnect()

        # If the system is configured to display the acquired frames to the user, ensures that the display frame rate
        # is valid and works with the managed camera's frame acquisition rate.
        if (display_frame_rate is not None and not isinstance(display_frame_rate, int)) or (
            isinstance(display_frame_rate, int)
            and (display_frame_rate <= 0 or display_frame_rate > self._camera.frame_rate)
        ):
            message = (
                f"Unable to configure the camera interface for the VideoSystem with id {self._system_id}. Encountered "
                f"an unsupported 'display_frame_rate' argument value {display_frame_rate} of type "
                f"{type(display_frame_rate).__name__}. The display frame rate override has to be None or a positive "
                f"integer that does not exceed the camera acquisition frame rate ({self._camera.frame_rate})."
            )
            console.error(error=TypeError, message=message)

        # Disables frame displaying on macOS as this OS does not support displaying frames outside the main thread.
        if display_frame_rate is not None and "darwin" in sys.platform:
            warnings.warn(
                message=(
                    f"Displaying frames is currently not supported for Apple Silicon devices. See ReadMe for details. "
                    f"Disabling frame display for the VideoSystem with id {self._system_id}."
                ),
                stacklevel=2,
            )
            display_frame_rate = None

        # Ensures that the display frame rate is stored as an integer and saves it to an attribute.
        self._display_frame_rate: int = display_frame_rate if display_frame_rate is not None else 0

        # Only adds the video saver if the user intends to save the acquired frames (as indicated by providing a valid
        # output directory).
        self._saver: VideoSaver | None = None
        if self._output_file is not None:
            # Validates the video saver configuration parameters:
            if not isinstance(gpu, int):
                message = (
                    f"Unable to configure the video saver for the VideoSystem with id {self._system_id}. Expected an "
                    f"integer as the 'gpu' argument value, but got {gpu} of type {type(gpu).__name__}."
                )
                console.error(error=TypeError, message=message)
            if video_encoder not in VideoEncoders:
                message = (
                    f"Unable to configure the video saver for the VideoSystem with id {self._system_id}. Encountered "
                    f"an unexpected 'video_encoder' argument value {video_encoder} of type "
                    f"{type(video_encoder).__name__}. Use one of the supported VideoEncoders enumeration members: "
                    f"{', '.join(tuple(VideoEncoders))}."
                )
                console.error(error=ValueError, message=message)
            if encoder_speed_preset not in EncoderSpeedPresets:
                message = (
                    f"Unable to configure the video saver for the VideoSystem with id {self._system_id}. Encountered "
                    f"an unexpected 'encoder_speed_preset' argument value {encoder_speed_preset} of type "
                    f"{type(encoder_speed_preset).__name__}. Use one of the supported EncoderSpeedPresets enumeration "
                    f"members: {', '.join([str(preset) for preset in tuple(EncoderSpeedPresets)])}."
                )
                console.error(error=ValueError, message=message)
            if output_pixel_format not in OutputPixelFormats:
                message = (
                    f"Unable to configure the video saver for the VideoSystem with id {self._system_id}. Encountered "
                    f"an unexpected 'output_pixel_format' argument value {output_pixel_format} of type "
                    f"{type(output_pixel_format).__name__}. Use one of the supported OutputPixelFormats enumeration "
                    f"members: {', '.join(tuple(OutputPixelFormats))}."
                )
                console.error(error=ValueError, message=message)
            if (
                not isinstance(quantization_parameter, int)
                or not -1 <= quantization_parameter <= _MAXIMUM_QUANTIZATION_VALUE
            ):
                message = (
                    f"Unable to configure the video saver for the VideoSystem with id {self._system_id}. Expected an "
                    f"integer between -1 and 51 as the 'quantization_parameter' argument value, but got "
                    f"{quantization_parameter} of type {type(quantization_parameter).__name__}."
                )
                console.error(error=TypeError, message=message)

            # VideoSaver relies on the FFMPEG library to be available on the system Path. Ensures that FFMPEG is
            # available for this runtime.
            if not check_ffmpeg_availability():
                message = (
                    f"Unable to configure the video saver for the VideoSystem with id {self._system_id}. VideoSaver "
                    f"requires a third-party software, FFMPEG, to be available on the system's Path. Make sure FFMPEG "
                    f"is installed and callable from a Python shell. See https://www.ffmpeg.org/download.html for more "
                    f"information."
                )
                console.error(error=RuntimeError, message=message)

            # Since GPU encoding is currently only supported for NVIDIA GPUs, verifies that nvidia-smi is callable
            # for the host system. This is used as a proxy to determine whether the system has an Nvidia GPU:
            if gpu >= 0 and not check_gpu_availability():
                message = (
                    f"Unable to configure the video saver for the VideoSystem with id {self._system_id}. The saver is "
                    f"configured to use the GPU video encoder, which currently only supports NVIDIA GPUs. Calling "
                    f"'nvidia-smi' to verify the presence of NVIDIA GPUs did not run successfully, indicating "
                    f"that there are no available NVIDIA GPUs on the host system. Use a CPU encoder or make sure "
                    f"nvidia-smi is callable from a Python shell."
                )
                console.error(error=RuntimeError, message=message)

            # Instantiates the VideoSaver object
            self._saver = VideoSaver(
                system_id=int(system_id),
                output_file=self._output_file,
                frame_height=self._camera.frame_height,
                frame_width=self._camera.frame_width,
                frame_rate=self._camera.frame_rate,
                gpu=gpu,
                video_encoder=video_encoder,
                encoder_speed_preset=encoder_speed_preset,
                input_pixel_format=self._camera.pixel_color_format,
                output_pixel_format=output_pixel_format,
                quantization_parameter=quantization_parameter,
            )

        # Sets up the assets used to manage acquisition and saver processes. The assets are configured during the
        # start() method runtime, most of them are initialized to placeholder values here.
        self._logger_queue: MPQueue = data_logger.input_queue  # type: ignore[type-arg]
        self._saver_queue: MPQueue = self._mp_manager.Queue()  # type: ignore[type-arg, assignment]
        self._terminator_array: SharedMemoryArray | None = None
        self._producer_process: Process | None = None
        self._consumer_process: Process | None = None
        self._watchdog_thread: Thread | None = None

    def __del__(self) -> None:
        """Releases all reserved resources before the instance is garbage-collected."""
        self.stop()
        self._mp_manager.shutdown()

    def __repr__(self) -> str:
        """Returns the string representation of the VideoSystem instance."""
        return (
            f"VideoSystem(system_id={self._system_id}, started={self._started}, "
            f"camera={type(self._camera).__name__!s}, frame_saving={self._saver is not None})"
        )

    def start(self) -> None:
        """Starts the instance's producer (camera interface) and consumer (video saver) processes and begins acquiring
        camera frames.

        Notes:
            Calling this method does not enable saving camera frames to non-volatile memory. To enable saving camera
            frames, call the start_frame_saving() method.

        Raises:
            RuntimeError: If starting the consumer or producer processes stalls or fails.
        """
        # Prevents restarting an already started VideoSystem instance.
        if self._started:
            return

        # This timer is used to forcibly terminate processes that stall at initialization.
        initialization_timer = PrecisionTimer(precision="s")

        # Instantiates a SharedMemoryArray used to control the runtime of the child processes.
        # Index 0 (element 1) is used to issue the global process termination command.
        # Index 1 (element 2) is used to flexibly enable or disable saving camera frames.
        # Index 2 (element 3) is used to track the producer process initialization status.
        # Index 3 (element 4) is used to track the consumer process initialization status.
        self._terminator_array = SharedMemoryArray.create_array(
            name=f"{self._system_id}_terminator_array",  # Uses class id with an additional specifier
            prototype=np.zeros(shape=4, dtype=np.uint8),
            exists_ok=True,  # Automatically recreates the buffer if it already exists
        )

        # Only starts the consumer process if the managed camera is configured to save frames.
        if self._saver is not None:
            self._consumer_process = Process(
                target=self._frame_saving_loop,
                args=(
                    self._system_id,
                    self._saver,
                    self._saver_queue,
                    self._logger_queue,
                    self._terminator_array,
                ),
                daemon=True,
            )
            self._consumer_process.start()

        # Starts the producer process
        self._producer_process = Process(
            target=self._frame_production_loop,
            args=(
                self._system_id,
                self._camera,
                self._display_frame_rate,
                self._saver_queue,
                self._logger_queue,
                self._terminator_array,
            ),
            daemon=True,
        )
        self._producer_process.start()

        # Connects to the shared memory array to receive control signals. It is important for this to be done after
        # both processes have been started.
        self._terminator_array.connect()
        # Ensures the buffer is destroyed if the instance is garbage-collected to prevent memory leaks.
        self._terminator_array.enable_buffer_destruction()

        # Waits for the processes to report that they have been successfully initialized.
        initialization_timer.reset()
        while self._terminator_array[2] != 1 and (
            self._consumer_process is not None and self._terminator_array[3] != 1
        ):  # pragma: no cover
            # If the processes take too long to initialize or die, raises an error.
            error = False
            message: str = ""  # Pre-initialization to appease mypy
            if (
                initialization_timer.elapsed > _PROCESS_INITIALIZATION_TIME and self._terminator_array[2] != 1
            ) or not self._producer_process.is_alive():
                message = (
                    f"Unable to start the VideoSystem with id {self._system_id}. The producer process has "
                    f"unexpectedly shut down or stalled for more than {_PROCESS_INITIALIZATION_TIME} seconds "
                    f"during initialization. This likely indicates a problem with the camera interface instance "
                    f"managed by the process."
                )
                error = True
            elif self._consumer_process is not None and (
                (initialization_timer.elapsed > _PROCESS_INITIALIZATION_TIME and self._terminator_array[3] != 1)
                or not self._consumer_process.is_alive()
            ):
                message = (
                    f"Unable to start the VideoSystem with id {self._system_id}. The consumer process has "
                    f"unexpectedly shut down or stalled for more than {_PROCESS_INITIALIZATION_TIME} seconds "
                    f"during initialization. This likely indicates a problem with the VideoSaver instance managed "
                    f"by the process."
                )
                error = True

            # Reclaims all committed resources before terminating with an error.
            if error:
                # Emits the process termination command
                self._terminator_array[0] = 1

                # Waits for any active processes to finish their execution
                if self._consumer_process is not None:
                    self._consumer_process.join()
                self._producer_process.join()

                # Disconnects from and destroys the shared memory array buffer
                self._terminator_array.disconnect()
                self._terminator_array.destroy()

                console.error(error=RuntimeError, message=message)

        # Creates and starts the watchdog thread
        self._watchdog_thread = Thread(target=self._watchdog, daemon=True)
        self._watchdog_thread.start()

        # Sets the _started flag, which also activates the watchdog monitoring.
        self._started = True

    def stop(self) -> None:
        """Stops the instance's producer (camera interface) and consumer (video saver) processes and releases all
        reserved resources.

        Notes:
            The consumer process is kept alive until all frames buffered to the saver_queue are saved. However, if the
            saver_queue does not become empty within 10 minutes from calling this method, it forcibly terminates the
            consumer process and discards any unprocessed data.
        """
        # Prevents stopping an already stopped VideoSystem instance.
        if not self._started or self._terminator_array is None:
            return

        # This timer is used to forcibly terminate the process that gets stuck in the shutdown sequence.
        shutdown_timer = PrecisionTimer(precision="s")

        # This inactivates the watchdog thread monitoring, ensuring it does not err when the processes are terminated.
        self._started = False

        # Emits the process shutdown signal.
        self._terminator_array[0] = 1

        # Delays for 2 seconds to allow the consumer process to terminate its runtime
        shutdown_timer.delay(delay=2, allow_sleep=True, block=False)

        # Waits until the saver_queue is empty. This is aborted if the shutdown stalls at this step for 10+ minutes.
        while not self._saver_queue.empty():
            if shutdown_timer.elapsed > _PROCESS_SHUTDOWN_TIME:
                break

        # Joins the producer and consumer processes
        if self._producer_process is not None:
            self._producer_process.join(timeout=20)
        if self._consumer_process is not None:
            self._consumer_process.join(timeout=20)

        # Joins the watchdog thread
        if self._watchdog_thread is not None:
            self._watchdog_thread.join(timeout=20)

        # Disconnects from and destroys the terminator array buffer.
        self._terminator_array.disconnect()
        self._terminator_array.destroy()

    @staticmethod
    def _frame_display_loop(
        display_queue: Queue,  # type: ignore[type-arg]
        system_id: np.uint8,
    ) -> None:  # pragma: no cover
        """Continuously fetches frame images from the display_queue and displays them via the OpenCV's imshow()
        function.

        Notes:
            This method runs in a thread as part of the _produce_images_loop() runtime in the producer Process.

        Args:
            display_queue: The multithreading Queue that buffers the grabbed camera frames until they are displayed.
            system_id: The unique identifier of the VideoSystem that generated the visualized images.
        """
        # Initializes the display window using the 'normal' mode to support user-controlled resizing.
        window_name = f"VideoSystem {system_id} Frames."
        cv2.namedWindow(winname=window_name, flags=cv2.WINDOW_NORMAL)

        # Runs until manually terminated by the user through the GUI or programmatically through the thread kill
        # argument.
        while True:
            # It is safe to fetch the frames in the blocking mode, since the loop is terminated by passing 'None'
            # through the queue
            frame = display_queue.get()

            # Programmatic termination is done by passing a non-numpy-array input through the queue
            if not isinstance(frame, np.ndarray):
                display_queue.task_done()  # If the thread is terminated, ensures join() works as expected
                break

            # Displays the image using the window created above
            cv2.imshow(winname=window_name, mat=frame)

            # Manual termination is done through the window GUI
            escape_key = 27  # The code for the ESC key in ASCII
            if cv2.waitKey(1) & 0xFF == escape_key:
                display_queue.task_done()  # If the thread is terminated, ensures join() works as expected
                break

            # Ensures that each queue get() call is paired with a task_done() call once display cycle is over
            display_queue.task_done()

        # Cleans up after runtime by destroying the window. Specifically targets the window created by this thread to
        # avoid interfering with any other windows.
        cv2.destroyWindow(winname=window_name)

    @staticmethod
    def _frame_production_loop(
        system_id: np.uint8,
        camera: OpenCVCamera | HarvestersCamera | MockCamera,
        display_frame_rate: int,
        saver_queue: MPQueue,  # type: ignore[type-arg]
        logger_queue: MPQueue,  # type: ignore[type-arg]
        terminator_array: SharedMemoryArray,
    ) -> None:  # pragma: no cover
        """Continuously grabs frames from the managed camera and queues them up to be saved by the consumer process.

        If the VideoSystem instance is configured to display acquired frame data, this method also uses a separate
        thread to render the acquired frames into and display them to the user in real time.

        Notes:
            This method should be executed by the producer Process. It is not intended to be executed by the main
            process where the VideoSystem is instantiated.

        Args:
            system_id: The unique identifier code of the caller VideoSystem instance. This is used to identify the
                VideoSystem in data log entries.
            camera: The camera interface instance for the camera from which to acquire frames.
            display_frame_rate: The desired rate, in frames per second, at which to display (visualize) the acquired
                frame stream to the user. Setting this argument to 0 disables frame display functionality.
            saver_queue: The multiprocessing Queue that buffers and pipes acquired frames to the consumer process.
            logger_queue: The multiprocessing Queue that buffers and pipes log entries to the DataLogger's logger
                process.
            terminator_array: A SharedMemoryArray instance used to control the runtime behavior of the process
                and terminate it during the global shutdown.
        """
        # Connects to the terminator array.
        terminator_array.connect()

        # Creates a timer that time-stamps acquired frames.
        frame_timer: PrecisionTimer = PrecisionTimer("us")

        # Constructs a timezone-aware stamp using UTC time. This creates a reference point for all future time
        # readouts.
        onset: NDArray[np.uint8] = get_timestamp(output_format=TimestampFormats.BYTES)  # type: ignore[assignment]
        frame_timer.reset()  # Immediately resets the stamp timer to make it as close as possible to the onset time

        # Sends the onset data to the logger queue. The acquisition_time of 0 is universally interpreted as the timer
        # onset.
        logger_queue.put(LogPackage(source_id=system_id, acquisition_time=np.uint64(0), serialized_data=onset))

        # If the camera is configured to display frames, creates a worker thread and a queue object that handles
        # displaying the frames.
        show_time: float | None = None
        show_timer: PrecisionTimer | None = None
        display_queue: Queue | None = None  # type: ignore[type-arg]
        display_thread: Thread | None = None
        if display_frame_rate > 0:
            # Creates the queue and thread for displaying camera frames
            display_queue = Queue()
            display_thread = Thread(target=VideoSystem._frame_display_loop, args=(display_queue, system_id))
            display_thread.start()

            # Converts the frame display rate from frames per second to microseconds per frame. This gives the delay
            # between displaying any two consecutive frames, which is used to limit how frequently the displayed image
            # updates.
            show_time = convert_time(time=1 / display_frame_rate, from_units="s", to_units="us", as_float=True)
            show_timer = PrecisionTimer("us")

        camera.connect()  # Connects to the hardware of the camera.

        # Indicates that the camera interface has started successfully.
        terminator_array[2] = 1

        try:
            # The loop runs until the VideoSystem is terminated by setting the first element (index 0) of the array to 1
            while not terminator_array[0]:
                # Grabs the first available frame as a numpy array. For Harvesters and Mock interfaces, this method
                # blocks until the frame is available if it is called too early. For OpenCV interface, this method
                # returns the same frame as grabbed during the previous call.
                frame = camera.grab_frame()
                frame_stamp = frame_timer.elapsed  # Generates the time-stamp for the acquired frame

                # If the camera is configured to display acquired frames, queues each frame to be displayed. The rate
                # at which the frames are displayed does not have to match the rate at which they are acquired.
                if display_queue is not None and show_timer.elapsed >= show_time:  # type: ignore[union-attr, operator]
                    # Resets the display timer
                    show_timer.reset()  # type: ignore[union-attr]
                    display_queue.put(frame)

                # If frame saving is enabled, sends the acquired frame data and the acquisition timestamp to the
                # consumer (video saver) process.
                if terminator_array[1] == 1:
                    saver_queue.put((frame, frame_stamp))

        # If an unknown and unhandled exception occurs, prints and flushes the exception message to the terminal
        # before re-raising the exception to terminate the process.
        except Exception as e:
            sys.stderr.write(str(e))
            sys.stderr.flush()
            raise

        # Ensures that local assets are always properly terminated
        finally:
            # Releases camera and shared memory assets.
            terminator_array.disconnect()
            camera.disconnect()

            # Terminates the display thread
            if display_queue is not None:
                display_queue.put(None)

            # Waits for the thread to close
            if display_thread is not None:
                display_thread.join()

    @staticmethod
    def _frame_saving_loop(
        system_id: np.uint8,
        saver: VideoSaver,
        saver_queue: MPQueue,  # type: ignore[type-arg]
        logger_queue: MPQueue,  # type: ignore[type-arg]
        terminator_array: SharedMemoryArray,
    ) -> None:  # pragma: no cover
        """Continuously grabs the frames from the image_queue and saves them as an .mp4 video file.

        This method also logs the acquisition time for each saved frame via the logger_queue instance.

        Notes:
            This method should be executed by the consumer Process. It is not intended to be executed by the main
            process where the VideoSystem is instantiated.

            This method's main loop is kept alive until the image_queue is empty. This is an intentional security
            feature that ensures all buffered images are processed before the saver is terminated. To override this
            behavior, you will need to use the process kill command, but it is strongly advised not to tamper
            with this feature.

        Args:
            system_id: The unique identifier code of the caller VideoSystem instance. This is used to identify the
                VideoSystem in data log entries.
            saver: The VideoSaver instance to use for saving the input frames as an .mp4 video file.
            saver_queue: The multiprocessing Queue that buffers and pipes acquired frames to the consumer process.
            logger_queue: The multiprocessing Queue that buffers and pipes log entries to the DataLogger's logger
                process.
            terminator_array: A SharedMemoryArray instance used to control the runtime behavior of the process
                and terminate it during the global shutdown.
        """
        # Connects to the terminator array used to manage the loop runtime.
        terminator_array.connect()

        # Initializes the FFMPEG encoder process.
        saver.start()

        # Indicates that the video saver has started successfully.
        terminator_array[3] = 1

        # Pre-creates the placeholder array used to log frame acquisition timestamps.
        data_placeholder = np.array([], dtype=np.uint8)

        try:
            # This loop runs until the global shutdown command is issued (via the variable under index 0) and until the
            # image_queue is empty.
            while not terminator_array[0] or not saver_queue.empty():
                # Grabs the frame data and its acquisition timestamp from the queue
                try:
                    frame: NDArray[np.integer[Any]]
                    frame_time: int
                    frame, frame_time = saver_queue.get_nowait()
                except Empty:
                    # Cycles the loop if the queue is empty
                    continue

                # Sends the frame to be saved by the saver
                saver.save_frame(frame)

                # Logs the saved frame's acquisition timestamp
                logger_queue.put(
                    LogPackage(
                        system_id,
                        acquisition_time=np.uint64(frame_time),
                        serialized_data=np.array(object=data_placeholder, dtype=np.uint8),
                    )
                )

        # If an unknown and unhandled exception occurs, prints and flushes the exception message to the terminal
        # before re-raising the exception to terminate the process.
        except Exception as e:
            sys.stderr.write(str(e))
            sys.stderr.flush()
            raise

        # Ensures that local assets are always properly terminated
        finally:
            # Disconnects from the shared memory array
            terminator_array.disconnect()

            # Stops the encoder process
            saver.stop()

    def _watchdog(self) -> None:  # pragma: no cover
        """This method is used by the watchdog thread to ensure that the producer and consumer processes are alive
        during runtime.

        This method raises RuntimeErrors if it detects that any of the processes has prematurely shut down. It verifies
        the process state in 20-millisecond cycles and releases the GIL between state verifications.

        Notes:
            If the method detects that the consumer or producer process has terminated prematurely, it carries out the
            necessary resource cleanup steps before raising the error and terminating the overall runtime.
        """
        # Initializes the timer used to space out the process state checks.
        timer = PrecisionTimer(precision="ms")

        # The watchdog function runs until the global shutdown signal is emitted.
        while self._terminator_array is not None and not self._terminator_array[0]:
            # Checks process state every 20 ms. Releases the GIL while waiting.
            timer.delay(delay=20, allow_sleep=True, block=False)

            # The watchdog functionality only kicks-in after the VideoSystem has been started
            if not self._started:
                continue

            # Checks if the producer is alive
            error = False
            producer = False
            if self._producer_process is not None and not self._producer_process.is_alive():
                error = True
                producer = True

            # Checks if the consumer is alive
            if self._consumer_process is not None and not self._consumer_process.is_alive():
                error = True

            # If either consumer or producer is dead, ensures proper resource reclamation before terminating with an
            # error
            if error:
                # Reclaims all committed resources before terminating with an error.
                self._terminator_array[0] = 1

                # If the consumer process is alive, gives it time to finish processing any remaining frames.
                while (
                    self._consumer_process is not None
                    and not self._saver_queue.empty()
                    and self._consumer_process.is_alive()
                ):
                    if timer.elapsed > _PROCESS_SHUTDOWN_TIME * 1000:
                        break

                # Joins all processes
                if self._consumer_process is not None:
                    self._consumer_process.join()
                if self._producer_process is not None:
                    self._producer_process.join()

                # Disconnects from the shared memory array and destroys the shared memory buffer.
                if self._terminator_array is not None:
                    self._terminator_array.disconnect()
                    self._terminator_array.destroy()

                # The code above is equivalent to stopping the instance
                self._started = False

                # Raises the error.
                if producer:
                    message = (
                        f"The producer process for the VideoSystem with id {self._system_id} has been prematurely "
                        f"shut down. This likely indicates that the process has encountered a runtime error that "
                        f"terminated the process."
                    )
                    console.error(message=message, error=RuntimeError)

                else:
                    message = (
                        f"The consumer process for the VideoSystem with id {self._system_id} has been prematurely "
                        f"shut down. This likely indicates that the process has encountered a runtime error that "
                        f"terminated the process."
                    )
                    console.error(message=message, error=RuntimeError)

    def start_frame_saving(self) -> None:
        """Enables saving acquired camera frames to disk as an .mp4 video file."""
        if self._started and self._terminator_array is not None:
            self._terminator_array[1] = 1

    def stop_frame_saving(self) -> None:
        """Disables saving acquired camera frames to disk as an .mp4 video file.

        Notes:
            Calling this method does not stop the frame acquisition process. It only prevents the acquired frames from
            being sent to the consumer process, which prevents them from being saved to disk.
        """
        if self._started and self._terminator_array is not None:
            self._terminator_array[1] = 0

    @property
    def video_file_path(self) -> Path | None:
        """Returns the path to the output video file if the instance is configured to save acquired camera frames and
        None otherwise.
        """
        return self._output_file if self._saver is not None else None

    @property
    def started(self) -> bool:
        """Returns True if the system has been started and has active producer and (optionally) consumer processes."""
        return self._started

    @property
    def system_id(self) -> np.uint8:
        """Returns the unique identifier code assigned to the VideoSystem instance."""
        return self._system_id


def _process_frame_message_batch(log_path: Path, file_names: list[str], onset_us: np.uint64) -> list[np.uint64]:
    """Processes the target batch of VideoSystem-generated messages stored in the .npz log file.

    This worker function is used by the _extract_camera_timestamps() function to process multiple message batches in
    parallel to speed up the overall camera timestamp data processing.

    Args:
        log_path: The path to the processed .npz log file.
        file_names: The names of the individual message .npy files stored in the target archive.
        onset_us: The onset of the frame data acquisition, in microseconds elapsed since UTC epoch onset.

    Returns:
        The list of frame acquisition timestamps for all frames whose messages have been processed as part of the
        batch, stored as microseconds since UTC epoch onset.
    """
    # Opens the processed log archive using memory mapping. If frame processing is performed in parallel, all processes
    # interact with the archive concurrently.
    with np.load(log_path, allow_pickle=False, fix_imports=False, mmap_mode="r") as archive:
        frame_timestamps = []

        # Loops over the batch of frame messages and extracts frame acquisition timestamps.
        for item in file_names:
            message = archive[item]

            # Frame timestamp messages do not have a payload, they only contain the source ID and the acquisition
            # timestamp. This gives them the length of 9 bytes.
            if len(message) == _FRAME_TIMESTAMP_LOG_MESSAGE_SIZE:
                # Extracts the number of microseconds elapsed since acquisition onset and uses it to calculate the
                # global timestamp for the message, in microseconds since UTC epoch onset.
                elapsed_microseconds = message[1:9].view(np.uint64).item()
                frame_timestamps.append(onset_us + elapsed_microseconds)

    return frame_timestamps


def extract_logged_camera_timestamps(
    log_path: Path,
    n_workers: int = -1,
) -> tuple[np.uint64, ...]:
    """Extracts the video camera frame acquisition timestamps from the target .npz log file generated by a VideoSystem
    instance during runtime.

    This function reads the '.npz' archive generated by the DataLogger 'compress_logs' method for a VideoSystem
    instance and, if the system saved any frames acquired by the managed camera, extracts the tuple of frame timestamps.
    The order of timestamps in the tuple is sequential and matches the order in which the frames were appended to the
    .mp4 video file.

    Notes:
        The timestamps are given as microseconds elapsed since the UTC epoch onset.

        If the target .npz archive contains fewer than 2000 messages, the processing is carried out sequentially
        regardless of the specified worker-count.

    Args:
        log_path: The path to the .npz log file that stores the logged data generated by the VideoSystem
            instance during runtime.
        n_workers: The number of parallel worker processes (CPU cores) to use for processing. Setting this to a value
            below 1 uses all available CPU cores. Setting this to a value of 1 conducts the processing sequentially.

    Returns:
        A tuple that stores the frame acquisition timestamps. Each timestamp is stored as the number of microseconds
        elapsed since the UTC epoch onset.

    Raises:
        ValueError: If the target .npz archive does not exist.
    """
    # Ensures that the target .npz log archive exists.
    if not log_path.exists() or log_path.suffix != ".npz" or not log_path.is_file():
        error_message = (
            f"Unable to extract camera frame timestamp data from the log file {log_path}, as it does not exist or does "
            f"not point to a valid .npz archive."
        )
        console.error(message=error_message, error=ValueError)

    # Memory-maps the processed archive to conserve RAM. The first processing pass is designed to find the onset
    # timestamp value.
    with np.load(log_path, allow_pickle=False, fix_imports=False, mmap_mode="r") as archive:
        # Locates the logging onset timestamp. The onset is used to convert the relative timestamps for logged frame
        # data into absolute UTC timestamps. Originally, all timestamps other than onset are stored as elapsed time in
        # microseconds relative to the onset timestamp.
        onset_us = np.uint64(0)
        timestamp_offset = 0
        message_list = list(archive.files)
        for number, item in enumerate(message_list):
            message: NDArray[np.uint8] = archive[item]  # Extracts message payload from the compressed .npy file

            # Recovers the uint64 timestamp value from each message. The timestamp occupies 8 bytes of each logged
            # message starting at index 1. If the timestamp value is 0, the message contains the onset timestamp value
            # stored as an 8-byte payload. Index 0 stores the source ID (uint8 value).
            timestamp_value = message[1:9].view(np.uint64).item()
            if timestamp_value == 0:
                # Extracts the byte-serialized UTC timestamp stored as microseconds since epoch onset.
                onset_us = np.uint64(message[9:].view(np.int64).item())

                # Breaks the loop once the onset is found. Generally, the onset is expected to be found very early into
                # the loop.
                timestamp_offset = number  # Records the item number at which the onset value was found.
                break

    # Builds the list of files to process after discovering the timestamp (the list of remaining messages)
    messages_to_process = message_list[timestamp_offset + 1 :]

    # If there are no leftover messages to process, return an empty tuple
    if not messages_to_process:
        return ()

    # Small archives are processed sequentially to avoid the unnecessary overhead of setting up the multiprocessing
    # runtime. This is also done for large files if the user explicitly requests to use a single worker process.
    if n_workers == 1 or len(messages_to_process) < _MINIMUM_LOG_SIZE_FOR_PARALLELIZATION:
        return tuple(_process_frame_message_batch(log_path, messages_to_process, onset_us))

    # If the user enabled using all available cores, configures the runtime to use all available CPUs
    if n_workers < 0:
        n_workers = cpu_count()

    # Creates batches of messages to process during runtime. Uses a fairly high batch multiplier to create many smaller
    # batches, which leads to a measurable increase in the processing speed, especially for large archives. The optimal
    # multiplier value (4) was determined experimentally.
    batches = []
    batch_indices = []  # Keeps track of batch order
    for i, batch in enumerate(chunk_iterable(messages_to_process, n_workers * 4)):
        if batch:
            batches.append((log_path, list(batch), onset_us))
            batch_indices.append(i)

    # Processes batches using ProcessPoolExecutor
    with ProcessPoolExecutor(max_workers=n_workers) as executor:
        # Submits all tasks
        future_to_index = {
            executor.submit(_process_frame_message_batch, *batch_args): idx
            for idx, batch_args in zip(batch_indices, batches, strict=False)
        }

        # Collects results while maintaining frame order. This also propagates processing errors to the caller process.
        results: list[list[np.uint64] | None] = [None] * len(batches)

        # Creates a progress bar for batch processing
        with tqdm(total=len(batches), desc="Extracting camera frame timestamps", unit="batch") as pbar:
            for future in as_completed(future_to_index):
                results[future_to_index[future]] = future.result()
                pbar.update(1)  # Updates the progress bar after each batch completes

    # Combines processing results in order
    all_timestamps: list[np.uint64] = []
    for batch_timestamps in results:
        # noinspection PyUnreachableCode
        if batch_timestamps is not None:  # Skips None results
            all_timestamps.extend(batch_timestamps)

    return tuple(all_timestamps)



"""This module provides the Command Line Interface (CLI) installed into the python's environment together with the
library.
"""

from pathlib import Path  # pragma: no cover

import click  # pragma: no cover
import numpy as np  # pragma: no cover
from ataraxis_base_utilities import LogLevel, console  # pragma: no cover
from ataraxis_data_structures import DataLogger, assemble_log_archives  # pragma: no cover

from .saver import (
    OutputPixelFormats,
    EncoderSpeedPresets,
    check_gpu_availability,
    check_ffmpeg_availability,
)  # pragma: no cover
from .camera import CameraInterfaces, add_cti_file, get_opencv_ids, get_harvesters_ids  # pragma: no cover
from .video_system import VideoSystem  # pragma: no cover

# Enables console output
console.enable()  # pragma: no cover

# Ensures that displayed CLICK help messages are formatted according to the lab standard.
CONTEXT_SETTINGS = {"max_content_width": 120}  # pragma: no cover


@click.group("axvs", context_settings=CONTEXT_SETTINGS)
def axvs_cli() -> None:  # pragma: no cover
    """This Command-Line Interface (CLI) functions as the entry-point for interfacing with all interactive components
    of the ataraxis-video-system (AXVS) library.
    """


@axvs_cli.command("cti")
@click.option(
    "-f",
    "--file-path",
    required=False,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=Path),
    help=(
        "The path to the CTI file that provides the GenTL Producer interface. It is recommended to use the "
        "file supplied by the camera vendor, but a general Producer, such as mvImpactAcquire, is also acceptable. "
        "See https://github.com/genicam/harvesters/blob/master/docs/INSTALL.rst for more details."
    ),
)
def set_cti_file(file_path: Path) -> None:  # pragma: no cover
    """Configures the library to use the input CTI file for all future runtimes involving GeniCam cameras.

    This library relies on the Harvesters library to interface with GeniCam-compatible cameras. In turn, the Harvesters
    library requires the GenTL Producer interface (.cti) file to discover and interface with compatible cameras. This
    command must be called at least once before calling all other CLIs and APIs that rely on the Harvesters library.
    """
    add_cti_file(cti_path=file_path)

    # Notifies the user that the CTI file has been successfully set.
    console.echo(f"AXVS CTI file: Set to {file_path}.", level=LogLevel.SUCCESS)


@axvs_cli.command("id")
def list_camera_indices() -> None:
    """Discovers all cameras compatible with the Opencv and Harvesters interfaces and prints their identification
    information.

    This command is primarily intended to be used during the initial system configuration to determine the positional
    indices of each camera in the list of all cameras discoverable by each supported interface. The discovered indices
    can then be used to initialize the VideoSystem instances to interface with the discovered cameras.
    """
    # Discovers compatible OpenCV cameras
    opencv_cameras = get_opencv_ids()

    # If no cameras are discovered, displays an error message and advances to Harvesters verification.
    if len(opencv_cameras) == 0:
        console.echo(message="No OpenCV-compatible cameras discovered.", level=LogLevel.ERROR)

    else:
        # Otherwise, lists the data for all discovered cameras.
        console.echo(
            message=(
                "Warning! Currently, it is impossible to resolve camera models or serial numbers through the "
                "OpenCV interface. It is recommended to check each discovered OpenCV camera via the 'axvs run' "
                "CLI command to precisely map the discovered camera indices to specific camera hardware."
            ),
            level=LogLevel.WARNING,
        )
        console.echo("Available OpenCV cameras:", level=LogLevel.SUCCESS)
        for num, camera_data in enumerate(opencv_cameras, start=1):
            console.echo(
                message=(
                    f"OpenCV camera {num}: index={camera_data.camera_index}, "
                    f"frame_height={camera_data.frame_height} pixels, frame_width={camera_data.frame_width} pixels, "
                    f"frame_rate={camera_data.acquisition_frame_rate} frames / second."
                )
            )
    try:
        harvesters_cameras = get_harvesters_ids()

        if len(harvesters_cameras) == 0:
            console.echo(message="No Harvesters-compatible cameras discovered.", level=LogLevel.ERROR)
            return

        # Note, Harvesters interface supports identifying the camera's model and serial number, which makes it easy to
        # mao discovered indices to physical hardware.
        console.echo("Available Harvesters cameras:", level=LogLevel.SUCCESS)
        for num, camera_data in enumerate(harvesters_cameras, start=1):
            console.echo(
                message=(
                    f"Harvesters camera {num}: index={camera_data.camera_index}, model={camera_data.model}, "
                    f"serial_code={camera_data.serial_number} frame_height={camera_data.frame_height} pixels,"
                    f"frame_width={camera_data.frame_width} pixels, "
                    f"frame_rate={camera_data.acquisition_frame_rate} frames / second."
                )
            )
    except Exception:
        console.echo(
            message=(
                "Unable to discover Harvesters-compatible cameras, as the library has not been provided with a CTI "
                "interface file. Use the 'axvs cti' command to set the path to the CTI file before calling this "
                "command."
            ),
            level=LogLevel.ERROR,
        )


@axvs_cli.command("check")
def check_requirements() -> None:  # pragma: no cover
    """Checks whether the host system meets the requirements for CPU and (optionally) GPU video encoding.

    This command allows checking whether the local system is set up correctly to support saving acquired camera frames
    as videos. As a minimum, this requires that the system has the FFMPEG library installed and available on the
    system's Path. Additionally, to support GPU (hardware) encoding, the system must have an Nvidia GPU. Note; the
    presence of the GPU is evaluated by calling the 'nvidia-smi' command, so it must also be installed on the local
    system alongside the GPU for the check to work as expected.
    """
    if not check_ffmpeg_availability():
        console.echo(
            message="Video saving requirements: Not met. Unable to access the FFMPEG library.", level=LogLevel.ERROR
        )
    elif not check_gpu_availability():
        console.echo(
            message=(
                "Video saving requirements: Partially met. The local system supports CPU video encoding via the "
                "FFMPEG library, but does not have an Nvidia GPU for GPU encoding."
            ),
            level=LogLevel.WARNING,
        )
    else:
        console.echo(
            message="Video saving requirements: Fully met. The system supports both CPU and GPU video encoding.",
            level=LogLevel.SUCCESS,
        )


@axvs_cli.command("run")
@click.option(
    "-i",
    "--interface",
    type=click.Choice(["mock", "harvesters", "opencv"]),
    default="mock",
    show_default=True,
    help="The camera interface to use for interacting with the camera hardware. It is recommended to use the "
    "'harvesters' interface for all GeniCam-compatible cameras and the 'opencv' interface for all other cameras.",
)
@click.option(
    "-c",
    "--camera-index",
    type=int,
    default=0,
    show_default=True,
    help="The index of the target camera in the list of all cameras discoverable through the chosen interface. This "
    "option allows selecting the desired camera if multiple are available on the host-system.",
)
@click.option(
    "-g",
    "--gpu-index",
    type=int,
    default=-1,
    show_default=True,
    help="The index of the GPU device to use for video encoding. Setting this option to a value below zero (default) "
    "forces the VideoSystem to use the CPU for encoding the videos. Note; GPU encoding currently requires an "
    "Nvidia GPU that supports hardware video encoding.",
)
@click.option(
    "-o",
    "--output-directory",
    required=True,
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, path_type=Path),
    help="The path to the output directory where to save the acquired camera frames as an .mp4 video file.",
)
@click.option(
    "-m",
    "--monochrome",
    is_flag=True,
    default=False,
    show_default=True,
    help="Determines whether the camera records frames in monochrome (grayscale) or colored spectrum.",
)
@click.option(
    "-w",
    "--width",
    type=int,
    default=600,
    show_default=True,
    help="The width of the camera frames to acquire, in pixels.",
)
@click.option(
    "-h",
    "--height",
    type=int,
    default=400,
    show_default=True,
    help="The height of the camera frames to acquire, in pixels.",
)
@click.option(
    "-f",
    "--frame-rate",
    type=int,
    default=30,
    show_default=True,
    help="The rate at which to acquire the frames, in frames per second.",
)
def live_run(
    interface: str,
    camera_index: int,
    gpu_index: int,
    output_directory: Path,
    width: int,
    height: int,
    frame_rate: int,
    *,
    monochrome: bool,
) -> None:  # pragma: no cover
    """Creates a VideoSystem instance using the input parameters and starts an interactive imaging session.

    This command allows testing various components of the VideoSystem by running an interactive session controlled via
    the terminal. Primarily, this CLI is designed to help with the initial identification and calibration of VideoSystem
    instances and does not support the full range of features offered through the VideoSystem class API.
    """
    # Initializes and starts the DataLogger instance
    logger = DataLogger(output_directory=Path(output_directory), instance_name="axvs_live_run")
    logger.start()

    # Uses command arguments to resolve VideoSystem configuration parameters
    if interface == "mock":
        camera_interface = CameraInterfaces.MOCK
    elif interface == "harvesters":
        camera_interface = CameraInterfaces.HARVESTERS
    else:
        camera_interface = CameraInterfaces.OPENCV

    # Initializes the VideoSystem system
    video_system = VideoSystem(
        system_id=np.uint8(111),
        data_logger=logger,
        output_directory=Path(output_directory),
        camera_interface=camera_interface,
        camera_index=camera_index,
        frame_width=width,
        frame_height=height,
        frame_rate=frame_rate,
        display_frame_rate=25,  # Statically sets the display rate to 25 fps.
        color=not monochrome,
        gpu=gpu_index,
        video_encoder="H264",  # Older H264 codec for compatibility with older hardware.
        encoder_speed_preset=EncoderSpeedPresets.FAST,  # Faster encoding speed for compatibility with older hardware.
        output_pixel_format=OutputPixelFormats.YUV420,  # Half-width chroma coding.
        quantization_parameter=15,  # Uses the instance's default parameter
    )

    # Starts the system by spawning child processes
    video_system.start()
    console.echo(message="Live VideoSystem: initialized and started (spawned child processes).", level=LogLevel.INFO)

    # Ensures that manual control instruction is only shown once
    once: bool = True
    # Ues terminal input to control the video system
    while video_system.started:
        if once:
            message = (
                "Enter 'q' to terminate system's runtime. Enter 'w' to start saving camera frames. "
                "Enter 's' to stop saving camera frames. Note, after termination, the system may stay alive for up "
                "to 600 seconds to finish saving buffered frame data."
            )
            console.echo(message=message, level=LogLevel.SUCCESS)
            once = False

        key = input("\nEnter command key:")
        if key.lower() == "q":
            message = "Terminating the VideoSystem..."
            console.echo(message)
            video_system.stop()
            logger.stop()
        elif key.lower() == "w":  # pragma: no cover
            message = "VideoSystem's camera frame saving: Started."
            console.echo(message)
            video_system.start_frame_saving()
        elif key.lower() == "s":  # pragma: no cover
            message = "VideoSystem's camera frame saving: Stopped."
            console.echo(message)
            video_system.stop_frame_saving()
        else:  # pragma: no cover
            message = (
                f"Unknown input key {key.lower()} encountered while interacting with the VideoSystem. Use 'q' to "
                f"terminate the runtime, 'w' to start saving frames, and 's' to stop saving frames."
            )
            console.echo(message, level=LogLevel.WARNING)
    video_system.stop()
    logger.stop()
    console.echo(
        message=f"VideoSystem: Terminated. Saved frames (if any) are available from the {output_directory} directory.",
        level=LogLevel.SUCCESS,
    )
    assemble_log_archives(log_directory=logger.output_directory, remove_sources=True, verbose=True)
